from functools import reduce
from six.moves import builtins

from pyrtl.rtllib import multipliers as mult

from ..wire import Const, WireVector
from ..corecircuits import as_wires, concat, select
from ..pyrtlexceptions import PyrtlError
from ..helperfuncs import formatted_str_to_val


class Matrix(object):
    ''' Class for making a Matrix using PyRTL.

    Provides the ability to perform different matrix operations.
    '''
    # Internally, this class uses a Python matrix of WireVectors.
    # So, a Matrix is represented as follows for a 2 x 2:
    # [[WireVector, WireVector], [WireVector, WireVector]]

    def __init__(self, rows, columns, bits, signed=False, value=None, max_bits=64):
        ''' Constructs a Matrix object.

        :param int rows: the number of rows in the matrix.  Must be greater than 0
        :param int columns: the number of columns in the matrix.  Must be greater than 0
        :param int bits: The amount of bits per wirevector. Must be greater than 0
        :param bool signed: Currently not supported (will be added in the future)
        :param (WireVector/list) value: The value you want to initialize the Matrix with.
            If a WireVector, must be of size `rows * columns * bits`. If a list, must have
            `rows` rows and `columns` columns, and every element must fit in `bits` size.
            If not given, the matrix initializes to 0
        :param int max_bits: The maximum number of bits each wirevector can have, even
            after operations like adding two matrices together results in larger
            resulting wirevectors
        :return: a constructed Matrix object
        '''
        if not isinstance(rows, int):
            raise PyrtlError('Rows must be of type int, instead "%s" '
                             'was passed of type %s' %
                             (str(rows), type(rows)))
        if rows <= 0:
            raise PyrtlError('Rows cannot be less than or equal to zero. '
                             'Rows value passed: %s' % str(rows))

        if not isinstance(columns, int):
            raise PyrtlError('Columns must be of type int, instead "%s" '
                             'was passed of type %s' %
                             (str(columns), type(columns)))
        if columns <= 0:
            raise PyrtlError('Columns cannot be less than or equal to zero. '
                             'Columns value passed: %s' % str(columns))

        if not isinstance(bits, int):
            raise PyrtlError('Bits must be of type int, instead "%s" '
                             'was passed of type %s' %
                             (str(bits), type(bits)))
        if bits <= 0:
            raise PyrtlError(
                'Bits cannot be negative or zero, '
                'instead "%s" was passed' % str(bits))

        if max_bits is not None:
            if bits > max_bits:
                bits = max_bits

        self._matrix = [[0 for _ in range(columns)] for _ in range(rows)]

        if value is None:
            for i in range(rows):
                for j in range(columns):
                    self._matrix[i][j] = Const(0)
        elif isinstance(value, WireVector):
            if value.bitwidth != bits * rows * columns:
                raise PyrtlError('Initialized bitwidth value does not match '
                                 'given value.bitwidth: %s, expected: %s'
                                 '' % (str(value.bitwidth),
                                       str(bits * rows * columns)))
            for i in range(rows):
                for j in range(columns):
                    start_index = (j * bits) + (i * columns * bits)
                    self._matrix[rows - i - 1][columns - j - 1] =\
                        as_wires(value[start_index:start_index + bits], bitwidth=bits)

        elif isinstance(value, list):
            if len(value) != rows or any(len(row) != columns for row in value):
                raise PyrtlError('Rows and columns mismatch\n'
                                 'Rows: %s, expected: %s\n'
                                 'Columns: %s, expected: %s'
                                 '' % (str(len(value)), str(rows),
                                       str(len(value[0])), str(columns)))
            for i in range(rows):
                for j in range(columns):
                    self._matrix[i][j] = as_wires(value[i][j], bitwidth=bits)

        else:
            raise PyrtlError('Initialized value must be of type WireVector or '
                             'list. Instead was passed value of type %s' % (type(value)))

        self.rows = rows
        self.columns = columns
        self._bits = bits
        self.bits = bits
        self.signed = False
        self.max_bits = max_bits

    @property
    def bits(self):
        ''' Gets the number of bits each value is allowed to hold.

        :return: an integer representing the number of bits
        '''
        return self._bits

    @bits.setter
    def bits(self, bits):
        ''' Sets the number of bits.

        :param int bits: The number of bits. Must be greater than 0

        Called automatically when bits is changed.
        NOTE: This function will truncate the most significant bits.
        '''
        if not isinstance(bits, int):
            raise PyrtlError('Bits must be of type int, instead "%s" '
                             'was passed of type %s' %
                             (str(bits), type(bits)))
        if bits <= 0:
            raise PyrtlError(
                'Bits cannot be negative or zero, '
                'instead "%s" was passed' % str(bits))

        self._bits = bits
        for i in range(self.rows):
            for j in range(self.columns):
                self._matrix[i][j] = self._matrix[i][j][:bits]

    def __len__(self):
        ''' Gets the output WireVector length.

        :return: an integer representing the output WireVector bitwidth

        Used with default `len()` function
        '''
        return self.bits * self.rows * self.columns

    def to_wirevector(self):
        ''' Outputs the PyRTL Matrix as a singular concatenated Wirevector.

        :return: a Wirevector representing the whole PyRTL matrix

        For instance, if we had a 2 x 1 matrix `[[wire_a, wire_b]]` it would
        return the concatenated wire: `wire = wire_a.wire_b`
        '''
        result = []

        for i in range(len(self._matrix)):
            for j in range(len(self._matrix[0])):
                result.append(as_wires(self[i, j], bitwidth=self.bits))

        return as_wires(concat(*result), bitwidth=len(self))

    def transpose(self):
        ''' Constructs the transpose of the matrix

        :return: a Matrix object representing the transpose
        '''
        result = Matrix(self.columns, self.rows, self.bits, max_bits=self.max_bits)
        for i in range(result.rows):
            for j in range(result.columns):
                result[i, j] = self[j, i]
        return result

    def __reversed__(self):
        ''' Constructs the reverse of matrix

        :return: a Matrix object representing the reverse

        Used with the reversed() method
        '''
        result = Matrix(self.rows, self.columns, self.bits, max_bits=self.max_bits)
        for i in range(self.rows):
            for j in range(self.columns):
                result[i, j] = self[self.rows - 1 - i, self.columns - 1 - j]
        return result

    def __getitem__(self, key):
        ''' Accessor for the matrix.

        :param (int/slice row, int/slice column) key: The key value to get
        :return: WireVector or Matrix containing the value of key

        Called when using square brackets ([]).

        Examples::

            int_matrix = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
            matrix = Matrix.Matrix(3, 3, 4, value=int_matrix)

            matrix[1] == [3, 4, 5]
            matrix[2, 0] == 6
            matrix[(2, 0)] = 6
            matrix[slice(0, 2), slice(0, 3)] == [[0, 1, 2], [3, 4, 5]]
            matrix[0:2, 0:3] == [[0, 1, 2], [3, 4, 5]]
            matrix[:2] == [[0, 1, 2], [3, 4, 5]]
            matrix[-1] == [6, 7, 8]
            matrix[-2:] == [[3, 4, 5], [6, 7, 8]]
        '''
        if isinstance(key, tuple):
            rows, columns = key

            # First set up proper slice
            if not isinstance(rows, slice):
                if not isinstance(rows, int):
                    raise PyrtlError('Rows must be of type int or slice, '
                                     'instead "%s" was passed of type %s' %
                                     (str(rows), type(rows)))
                if rows < 0:
                    rows = self.rows - abs(rows)
                    if rows < 0:
                        raise PyrtlError("Invalid bounds for rows. Max rows: %s, got: %s" % (
                            str(self.rows), str(rows)))
                rows = slice(rows, rows + 1, 1)

            if not isinstance(columns, slice):
                if not isinstance(columns, int):
                    raise PyrtlError('Columns must be of type int or slice, '
                                     'instead "%s" was passed of type %s' %
                                     (str(columns), type(columns)))
                if columns < 0:
                    columns = self.columns - abs(columns)
                    if columns < 0:
                        raise PyrtlError("Invalid bounds for columns. Max columns: %s, got: %s" % (
                            str(self.columns), str(columns)))
                columns = slice(columns, columns + 1, 1)

            if rows.start is None:
                rows = slice(0, rows.stop, rows.step)
            elif rows.start < 0:
                rows = slice(self.rows - abs(rows.start),
                             rows.stop, rows.step)

            if rows.stop is None:
                rows = slice(rows.start, self.rows, rows.step)
            elif rows.stop < 0:
                rows = slice(rows.start, self.rows - abs(rows.stop),
                             rows.step)

            rows = slice(rows.start, rows.stop, 1)

            if columns.start is None:
                columns = slice(0, columns.stop, columns.step)
            elif columns.start < 0:
                columns = slice(self.columns - abs(columns.start),
                                columns.stop, columns.step)

            if columns.stop is None:
                columns = slice(columns.start, self.columns, columns.step)
            elif columns.stop < 0:
                columns = slice(
                    columns.start, self.columns - abs(columns.stop),
                    columns.step)

            columns = slice(columns.start, columns.stop, 1)

            # Check slice bounds
            if rows.start > self.rows or rows.stop > self.rows \
                    or rows.start < 0 or rows.stop < 0:
                raise PyrtlError("Invalid bounds for rows. Max rows: %s, got: %s" % (
                    str(self.rows), str(rows.start) + ":" + str(rows.stop)))
            if columns.start > self.columns or columns.stop > self.columns \
                    or columns.start < 0 or columns.stop < 0:
                raise PyrtlError("Invalid bounds for columns. Max columns: %s, got: %s" % (
                    str(self.columns), str(columns.start) + ":" + str(columns.stop)))

            # If it's a single value we want to return a wirevector
            if rows.stop - rows.start == 1 and \
                    columns.stop - columns.start == 1:
                return as_wires(self._matrix[rows][0][columns][0],
                                bitwidth=self.bits)

            # Otherwise set up matrix and return that
            result = [[0 for _ in range(columns.stop - columns.start)]
                      for _ in range(rows.stop - rows.start)]

            for i in range(len(result)):
                for j in range(len(result[0])):
                    result[i][j] = self._matrix[i + rows.start][j + columns.start]
            return Matrix(len(result), len(result[0]), self._bits,
                          signed=self.signed, value=result, max_bits=self.max_bits)

        # Second case when we just want to get full row
        if isinstance(key, int):
            if key < 0:
                start = self.rows - abs(key)
                if start < 0:
                    raise PyrtlError('Index %d is out of bounds for '
                                     'matrix with %d rows' % (key, self.rows))
                key = slice(start, start + 1, None)
            else:
                key = slice(key, key + 1, None)
            return self[key, :]

        # Third case when we want multiple rows
        if isinstance(key, slice):
            return self[key, :]

        # Otherwise improper value was passed
        raise PyrtlError('Rows must be of type int or slice, '
                         'instead "%s" was passed of type %s' %
                         (str(key), type(key)))

    def __setitem__(self, key, value):
        ''' Mutator for the matrix.

        :param (slice/int rows, slice/int columns) key: The key value to set
        :param Wirevector/int/Matrix value: The value in which to set the key

        Called when setting a value using square brackets (e.g. `matrix[a, b] = value`).

        The value given will be truncated to match the bitwidth of all the elements
        in the matrix.
        '''

        if isinstance(key, tuple):
            rows, columns = key

            # First ensure that slices are correct
            if not isinstance(rows, slice):
                if not isinstance(rows, int):
                    raise PyrtlError('Rows must be of type int or slice, '
                                     'instead "%s" was passed of type %s' %
                                     (str(rows), type(rows)))
                rows = slice(rows, rows + 1, 1)

            if not isinstance(columns, slice):
                if not isinstance(columns, int):
                    raise PyrtlError('Columns must be of type int or slice, '
                                     'instead "%s" was passed of type %s' %
                                     (str(columns), type(columns)))
                columns = slice(columns, columns + 1, 1)

            if rows.start is None:
                rows = slice(0, rows.stop, rows.step)
            elif rows.start < 0:
                rows = slice(self.rows - abs(rows.start),
                             rows.stop, rows.step)

            if rows.stop is None:
                rows = slice(rows.start, self.rows, rows.step)
            elif rows.stop < 0:
                rows = slice(rows.start, self.rows - abs(rows.stop),
                             rows.step)

            if columns.start is None:
                columns = slice(0, columns.stop, columns.step)
            elif columns.start < 0:
                columns = slice(self.columns - abs(columns.start),
                                columns.stop, columns.step)

            if columns.stop is None:
                columns = slice(columns.start, self.columns, columns.step)
            elif columns.stop < 0:
                columns = slice(
                    columns.start, self.columns - abs(columns.stop),
                    columns.step)

            # Check Slice Bounds
            if rows.start > self.rows or rows.stop > self.rows \
                    or rows.start < 0 or rows.stop < 0:
                raise PyrtlError("Invalid bounds for rows. Max rows: %s, got: %s" % (
                    str(self.rows), str(rows.start) + ":" + str(rows.stop)))
            if columns.start > self.columns or columns.stop > self.columns \
                    or columns.start < 0 or columns.stop < 0:
                raise PyrtlError("Invalid bounds for columns. Max columns: %s, got: %s" % (
                    str(self.columns), str(columns.start) + ":" + str(columns.stop)))

            # First case when setting value to Matrix
            if isinstance(value, Matrix):
                if value.rows != (rows.stop - rows.start):
                    raise PyrtlError(
                        'Value rows mismatch. Expected Matrix '
                        'of rows "%s", instead recieved Matrix of rows "%s"' %
                        (str(rows.stop - rows.start), str(value.rows)))
                if value.columns != (columns.stop - columns.start):
                    raise PyrtlError(
                        'Value columns mismatch. Expected Matrix '
                        'of columns "%s", instead recieved Matrix of columns "%s"' %
                        (str(columns.stop - columns.start), str(value.columns)))

                for i in range(rows.stop - rows.start):
                    for j in range(columns.stop - columns.start):
                        self._matrix[rows.start + i][columns.start + j] =\
                            as_wires(value[i, j], bitwidth=self.bits)

            # Second case when setting value to wirevector
            elif isinstance(value, (int, WireVector)):
                if ((rows.stop - rows.start) != 1) or \
                        ((columns.stop - columns.start) != 1):
                    raise PyrtlError(
                        'Value mismatch: expected Matrix, instead received WireVector')
                self._matrix[rows.start][columns.start] = as_wires(value, bitwidth=self.bits)
            # Otherwise Error
            else:
                raise PyrtlError('Invalid value of type %s' % type(value))
        else:
            # Second case if we just want to set a full row
            if isinstance(key, int):
                if key < 0:
                    start = self.rows - abs(key)
                    if start < 0:
                        raise PyrtlError('Index %d is out of bounds for '
                                         'matrix with %d rows' % (key, self.rows))
                    key = slice(start, start + 1, None)
                else:
                    key = slice(key, key + 1, None)
                self[key, :] = value
            # Third case if we want to set full rows
            elif isinstance(key, slice):
                self[key, :] = value
            else:
                raise PyrtlError('Rows must be of type int or slice, '
                                 'instead "%s" was passed of type %s' %
                                 (str(key), type(key)))

    def copy(self):
        ''' Constructs a deep copy of the Matrix.

        :return: a Matrix copy
        '''
        return Matrix(self.rows, self.columns, self.bits,
                      value=self.to_wirevector(), max_bits=self.max_bits)

    def __iadd__(self, other):
        ''' Perform the in-place addition operation.

        :return: a Matrix object with the elementwise addition being preformed

        Is used with `a += b`. Performs an elementwise addition.
        '''
        new_value = (self + other)
        self._matrix = new_value._matrix
        self.bits = new_value._bits
        return self.copy()

    def __add__(self, other):
        ''' Perform the addition operation.

        :return: a Matrix object with the element wise addition being performed

        Is used with `a + b`. Performs an elementwise addition.
        '''
        if not isinstance(other, Matrix):
            raise PyrtlError('error: expecting a Matrix, '
                             'got %s instead' % type(other))

        if self.columns != other.columns:
            raise PyrtlError('error: columns mismatch. '
                             'Matrix a: %s columns, Matrix b: %s rows' %
                             (str(self.columns), str(other.columns)))
        elif self.rows != other.rows:
            raise PyrtlError('error: row mismatch. '
                             'Matrix a: %s columns, Matrix b: %s column' %
                             (str(self.rows), str(other.rows)))

        new_bits = self.bits
        if other.bits > new_bits:
            new_bits = other.bits

        result = Matrix(self.rows, self.columns, new_bits + 1, max_bits=self.max_bits)

        for i in range(result.rows):
            for j in range(result.columns):
                result[i, j] = self[i, j] + other[i, j]
        return result

    def __isub__(self, other):
        ''' Perform the inplace subtraction opperation.

        :Matrix other: the PyRTL Matrix to subtract
        :return: a Matrix object with the element wise subtraction being performed

        Is used with `a -= b`. Performs an elementwise subtraction.
        '''
        new_value = self - other
        self._matrix = new_value._matrix
        self._bits = new_value._bits
        return self.copy()

    def __sub__(self, other):
        ''' Perform the subtraction operation.

        :Matrix other: the PyRTL Matrix to subtract
        :return: a Matrix object with the elementwise subtraction being performed

        Is used with `a - b`. Performs an elementwise subtraction.

        Note: If using unsigned numbers, the result will be floored at 0.
        '''
        if not isinstance(other, Matrix):
            raise PyrtlError('error: expecting a Matrix, '
                             'got %s instead' % type(other))

        if self.columns != other.columns:
            raise PyrtlError('error: columns mismatch. '
                             'Matrix a: %s columns, Matrix b: %s rows' %
                             (str(self.columns), str(other.columns)))

        if self.rows != other.rows:
            raise PyrtlError('error: row mismatch. '
                             'Matrix a: %s columns, Matrix b: %s column' %
                             (str(self.rows), str(other.rows)))

        new_bits = self.bits
        if other.bits > new_bits:
            new_bits = other.bits

        result = Matrix(self.rows, self.columns, new_bits, max_bits=self.max_bits)

        for i in range(result.rows):
            for j in range(result.columns):
                if self.signed:
                    result[i, j] = self[i, j] - other[i, j]
                else:
                    result[i, j] = select(self[i, j] > other[i, j],
                                          self[i, j] - other[i, j],
                                          Const(0))

        return result

    def __imul__(self, other):
        ''' Perform the in-place multiplication operation.

        :param Matrix/Wirevector other: the Matrix or scalar to multiply
        :return: a Matrix object with the resulting multiplication operation being preformed

        Is used with `a *= b`. Performs an elementwise or scalar multiplication.
        '''
        new_value = self * other
        self._matrix = new_value._matrix
        self._bits = new_value._bits
        return self.copy()

    def __mul__(self, other):
        ''' Perform the elementwise or scalar multiplication operation.

        :param Matrix/Wirevector other: the Matrix to multiply
        :return: a Matrix object with the resulting multiplication operation being performed

        Is used with `a * b`.
        '''

        if isinstance(other, Matrix):
            if self.columns != other.columns:
                raise PyrtlError('error: columns mismatch. '
                                 'Martrix a: %s columns, Matrix b: %s rows' %
                                 (str(self.columns), str(other.columns)))
            if self.rows != other.rows:
                raise PyrtlError('error, row mismatch '
                                 'Martrix a: %s columns, Matrix b: %s column' %
                                 (str(self.rows), str(other.rows)))

            bits = self.bits + other.bits
        elif isinstance(other, WireVector):
            bits = self.bits + len(other)
        else:
            raise PyrtlError('Expecting a Matrix or WireVector '
                             'got %s instead' % type(other))

        result = Matrix(self.rows, self.columns, bits, max_bits=self.max_bits)

        for i in range(self.rows):
            for j in range(self.columns):
                if isinstance(other, Matrix):
                    result[i, j] = self[i, j] * other[i, j]
                else:
                    result[i, j] = self[i, j] * other
        return result

    def __imatmul__(self, other):
        ''' Performs the inplace matrix multiplication operation.

        :param Matrix other: the second matrix.
        :return: a PyRTL Matrix that contains the matrix multiplication product of this and other

        Is used with `a @= b`.

        Note: The matmul symbol (@) only works in python 3.5+. Otherwise you must
        call `__imatmul__(other)`.
        '''
        new_value = self.__matmul__(other)
        self.columns = new_value.columns
        self.rows = new_value.rows
        self._matrix = new_value._matrix
        self._bits = new_value._bits
        return self.copy()

    def __matmul__(self, other):
        ''' Performs the matrix multiplication operation.

        :param Matrix other: the second matrix.
        :return: a PyRTL Matrix that contains the matrix multiplication product of this and other

        Is used with `a @ b`.

        Note: The matmul symbol (@) only works in python 3.5+. Otherwise you must
        call `__matmul__(other)`.
        '''
        if not isinstance(other, Matrix):
            raise PyrtlError('error: expecting a Matrix, '
                             'got %s instead' % type(other))

        if self.columns != other.rows:
            raise PyrtlError('error: rows and columns mismatch. '
                             'Matrix a: %s columns, Matrix b: %s rows' %
                             (str(self.columns), str(other.rows)))

        result = Matrix(self.rows, other.columns,
                        self.columns * other.rows * (self.bits + other.bits),
                        max_bits=self.max_bits)

        for i in range(self.rows):
            for j in range(other.columns):
                for k in range(self.columns):
                    result[i, j] = mult.fused_multiply_adder(
                        self[i, k], other[k, j], result[i, j], signed=self.signed)

        return result

    def __ipow__(self, power):
        ''' Performs the matrix power operation.

        :param int power: the power to perform the matrix on
        :return: a PyRTL Matrix that contains the matrix power product

        Is used with `a **= b`.
        '''
        new_value = self ** power
        self._matrix = new_value._matrix
        self._bits = new_value._bits
        return self.copy()

    def __pow__(self, power):
        ''' Performs the matrix power operation.

        :param int power: the power to perform the matrix on
        :return: a PyRTL Matrix that contains the matrix power product

        Is used with `a ** b`.
        '''
        if not isinstance(power, int):
            raise PyrtlError('Unexpected power given. Type int expected, '
                             'but recieved type %s' % type(power))

        if self.rows != self.columns:
            raise PyrtlError("Matrix must be square")

        result = self.copy()

        # First case: return identity matrix
        if power == 0:
            for i in range(self.rows):
                for j in range(self.columns):
                    if i != j:
                        result[i, j] = Const(0)
                    else:
                        result[i, j] = Const(1)
            return result
        # Second case: do matrix multiplications
        if power >= 1:
            inputs = [result] * power

            def pow_2(first, second):
                return first.__matmul__(second)
            return reduce(pow_2, inputs)

        raise PyrtlError('Power must be greater than or equal to 0')

    def put(self, ind, v, mode='raise'):
        ''' Replace specified elements of the matrix with given values

        :param int/list[int]/tuple[int] ind: target indices
        :param int/list[int]/tuple[int]/Matrix row-vector v: values to place in
            matrix at target indices; if v is shorter than ind, it is repeated as necessary
        :param str mode: how out-of-bounds indices behave; 'raise' raises an
            error, 'wrap' wraps aoround, and 'clip' clips to the range

        Note that the index is on the flattened matrix.
        '''
        count = self.rows * self.columns
        if isinstance(ind, int):
            ind = (ind,)
        elif not isinstance(ind, (tuple, list)):
            raise PyrtlError("Expected int or list-like indices, got %s" % type(ind))

        if isinstance(v, int):
            v = (v,)

        if isinstance(v, (tuple, list)) and len(v) == 0:
            return
        elif isinstance(v, Matrix):
            if v.rows != 1:
                raise PyrtlError(
                    "Expected a row-vector matrix, instead got matrix with %d rows" % v.rows
                )

        if mode not in ['raise', 'wrap', 'clip']:
            raise PyrtlError(
                "Unexpected mode %s; allowable modes are 'raise', 'wrap', and 'clip'" % mode
            )

        def get_ix(ix):
            if ix < 0:
                ix = count - abs(ix)
            if ix < 0 or ix >= count:
                if mode == 'raise':
                    raise PyrtlError("index %d is out of bounds with size %d" % (ix, count))
                elif mode == 'wrap':
                    ix = ix % count
                elif mode == 'clip':
                    ix = 0 if ix < 0 else count - 1
            return ix

        def get_value(ix):
            if isinstance(v, (tuple, list)):
                if ix >= len(v):
                    return v[-1]  # if v is shorter than ind, repeat last as necessary
                return v[ix]
            elif isinstance(v, Matrix):
                if ix >= count:
                    return v[0, -1]
                return v[0, ix]

        for v_ix, mat_ix in enumerate(ind):
            mat_ix = get_ix(mat_ix)
            row = mat_ix // self.columns
            col = mat_ix % self.columns
            self[row, col] = get_value(v_ix)

    def reshape(self, *newshape, **order):
        ''' Create a matrix of the given shape from the current matrix.

        :param int/ints/tuple[int] newshape: shape of the matrix to return;
            if a single int, will result in a 1-D row-vector of that length;
            if a tuple, will use values for number of rows and cols. Can also
            be a varargs.
        :param str order: 'C' means to read from self using
            row-major order (C-style), and 'F' means to read from self
            using column-major order (Fortran-style).
        :return: A copy of the matrix with same data, with a new number of rows/cols

        One shape dimension in newshape can be -1; in this case, the value
        for that dimension is inferred from the other given dimension (if any)
        and the number of elements in the matrix.

        Examples::

            int_matrix = [[0, 1, 2, 3], [4, 5, 6, 7]]
            matrix = Matrix.Matrix(2, 4, 4, value=int_matrix)

            matrix.reshape(-1) == [[0, 1, 2, 3, 4, 5, 6, 7]]
            matrix.reshape(8) == [[0, 1, 2, 3, 4, 5, 6, 7]]
            matrix.reshape(1, 8) == [[0, 1, 2, 3, 4, 5, 6, 7]]
            matrix.reshape((1, 8)) == [[0, 1, 2, 3, 4, 5, 6, 7]]
            matrix.reshape((1, -1)) == [[0, 1, 2, 3, 4, 5, 6, 7]]

            matrix.reshape(4, 2) == [[0, 1], [2, 3], [4, 5], [6, 7]]
            matrix.reshape(-1, 2) == [[0, 1], [2, 3], [4, 5], [6, 7]]
            matrix.reshape(4, -1) == [[0, 1], [2, 3], [4, 5], [6, 7]]
        '''
        # python2 does not support named arguments after *args, so we use
        # **kwargs for 'order' and set the default here.
        order = order.get('order', 'C')
        count = self.rows * self.columns
        if isinstance(newshape, int):
            if newshape == -1:
                newshape = (1, count)
            else:
                newshape = (1, newshape)
        elif isinstance(newshape, tuple):
            if isinstance(newshape[0], tuple):
                newshape = newshape[0]
            if len(newshape) == 1:
                newshape = (1, newshape[0])
            if len(newshape) > 2:
                raise PyrtlError("length of newshape tuple must be <= 2")
            rows, cols = newshape
            if not isinstance(rows, int) or not isinstance(cols, int):
                raise PyrtlError(
                    "newshape dimensions must be integers, instead got %s" % type(newshape)
                )
            if rows == -1 and cols == -1:
                raise PyrtlError("Both dimensions in newshape cannot be -1")
            if rows == -1:
                rows = count // cols
                newshape = (rows, cols)
            elif cols == -1:
                cols = count // rows
                newshape = (rows, cols)
        else:
            raise PyrtlError(
                "newshape can be an integer or tuple of integers, not %s" % type(newshape)
            )

        rows, cols = newshape
        if rows * cols != count:
            raise PyrtlError(
                "Cannot reshape matrix of size %d into shape %s" % (count, str(newshape))
            )

        if order not in 'CF':
            raise PyrtlError(
                "Invalid order %s. Acceptable orders are 'C' (for row-major C-style order) "
                "and 'F' (for column-major Fortran-style order)." % order
            )

        value = [[0] * cols for _ in range(rows)]
        ix = 0
        if order == 'C':
            # Read and write in row-wise order
            for newr in range(rows):
                for newc in range(cols):
                    r = ix // self.columns
                    c = ix % self.columns
                    value[newr][newc] = self[r, c]
                    ix += 1
        else:
            # Read and write in column-wise order
            for newc in range(cols):
                for newr in range(rows):
                    r = ix % self.rows
                    c = ix // self.rows
                    value[newr][newc] = self[r, c]
                    ix += 1

        return Matrix(rows, cols, self.bits, self.signed, value, self.max_bits)

    def flatten(self, order='C'):
        ''' Flatten the matrix into a single row.

        :param str order: 'C' means row-major order (C-style), and
            'F' means column-major order (Fortran-style)
        :return: A copy of the matrix flattened in to a row vector matrix
        '''
        return self.reshape(self.rows * self.columns, order=order)


def multiply(first, second):
    ''' Perform the elementwise or scalar multiplication operation.

    :param Matrix first: first matrix
    :param Matrix/Wirevector second: second matrix
    :return: a Matrix object with the element wise or scaler multiplication being performed
    '''
    if not isinstance(first, Matrix):
        raise PyrtlError('error: expecting a Matrix, '
                         'got %s instead' % type(second))
    return first * second


def sum(matrix, axis=None, bits=None):
    ''' Returns the sum of all the values in a matrix

    :param Matrix/Wirevector matrix: the matrix to perform sum operation on.
        If it is a WireVector, it will return itself
    :param None/int axis: The axis to perform the operation on
        None refers to sum of all item. 0 is sum of column. 1 is sum of rows. Defaults to None
    :param int bits: The bits per value of the sum. Defaults to bits of old matrix
    :return: A wirevector or Matrix representing sum
    '''
    def sum_2(first, second):
        return first + second

    if isinstance(matrix, WireVector):
        return matrix

    if not isinstance(matrix, Matrix):
        raise PyrtlError('error: expecting a Matrix or Wirevector for matrix, '
                         'got %s instead' % type(matrix))

    if not isinstance(bits, int) and bits is not None:
        raise PyrtlError('error: expecting an int/None for bits, '
                         'got %s instead' % type(bits))

    if not isinstance(axis, int) and axis is not None:
        raise PyrtlError('error: expecting an int or None for axis, '
                         'got %s instead' % type(axis))

    if bits is None:
        bits = matrix.bits

    if bits <= 0:
        raise PyrtlError('error: bits cannot be negative or zero, '
                         'got %s instead' % bits)

    if axis is None:
        inputs = []
        for i in range(matrix.rows):
            for j in range(matrix.columns):
                inputs.append(matrix[i, j])
        return reduce(sum_2, inputs)

    if axis == 0:
        result = Matrix(1, matrix.columns, signed=matrix.signed, bits=bits)

        for i in range(matrix.columns):
            inputs = []
            for j in range(matrix.rows):
                inputs.append(matrix[j, i])
            result[0, i] = reduce(sum_2, inputs)
        return result

    if axis == 1:
        result = Matrix(1, matrix.rows, signed=matrix.signed, bits=bits)
        for i in range(matrix.rows):
            inputs = []
            for j in range(matrix.columns):
                inputs.append(matrix[i, j])
            result[0, i] = reduce(sum_2, inputs)
        return result

    raise PyrtlError('Axis invalid: expected (None, 0, or 1), got %s' % axis)


def min(matrix, axis=None, bits=None):
    ''' Returns the minimum value in a matrix.

    :param Matrix/Wirevector matrix: the matrix to perform min operation on.
        If it is a WireVector, it will return itself
    :param None/int axis: The axis to perform the operation on
        None refers to min of all item. 0 is min of column. 1 is min of rows. Defaults to None
    :param int bits: The bits per value of the min. Defaults to bits of old matrix
    :return: A WireVector or Matrix representing the min value
    '''
    def min_2(first, second):
        return select(first < second, first, second)

    if isinstance(matrix, WireVector):
        return matrix

    if not isinstance(matrix, Matrix):
        raise PyrtlError('error: expecting a Matrix or Wirevector for matrix, '
                         'got %s instead' % type(matrix))

    if not isinstance(bits, int) and bits is not None:
        raise PyrtlError('error: expecting an int/None for bits, '
                         'got %s instead' % type(bits))

    if not isinstance(axis, int) and axis is not None:
        raise PyrtlError('error: expecting an int or None for axis, '
                         'got %s instead' % type(axis))

    if bits is None:
        bits = matrix.bits

    if bits <= 0:
        raise PyrtlError('error: bits cannot be negative or zero, '
                         'got %s instead' % bits)

    if axis is None:
        inputs = []
        for i in range(matrix.rows):
            for j in range(matrix.columns):
                inputs.append(matrix[i, j])
        return reduce(min_2, inputs)

    if axis == 0:
        result = Matrix(1, matrix.columns, signed=matrix.signed, bits=bits)

        for i in range(matrix.columns):
            inputs = []
            for j in range(matrix.rows):
                inputs.append(matrix[j, i])
            result[0, i] = reduce(min_2, inputs)
        return result

    if axis == 1:
        result = Matrix(1, matrix.rows, signed=matrix.signed, bits=bits)
        for i in range(matrix.rows):
            inputs = []
            for j in range(matrix.columns):
                inputs.append(matrix[i, j])
            result[0, i] = reduce(min_2, inputs)
        return result

    raise PyrtlError('Axis invalid: expected (None, 0, or 1), got %s' % axis)


def max(matrix, axis=None, bits=None):
    ''' Returns the max value in a matrix.

    :param Matrix/Wirevector matrix: the matrix to perform max operation on.
        If it is a wirevector, it will return itself
    :param None/int axis: The axis to perform the operation on
        None refers to max of all items. 0 is max of the columns. 1 is max of rows.
        Defaults to None
    :param int bits: The bits per value of the max. Defaults to bits of old matrix
    :return: A WireVector or Matrix representing the max value
    '''
    def max_2(first, second):
        return select(first > second, first, second)

    if isinstance(matrix, WireVector):
        return matrix

    if not isinstance(matrix, Matrix):
        raise PyrtlError('error: expecting a Matrix or WireVector for matrix, '
                         'got %s instead' % type(matrix))

    if not isinstance(bits, int) and bits is not None:
        raise PyrtlError('error: expecting an int/None for bits, '
                         'got %s instead' % type(bits))

    if not isinstance(axis, int) and axis is not None:
        raise PyrtlError('error: expecting an int or None for axis, '
                         'got %s instead' % type(axis))

    if bits is None:
        bits = matrix.bits

    if bits <= 0:
        raise PyrtlError('error: bits cannot be negative or zero, '
                         'got %s instead' % bits)

    if axis is None:
        inputs = []
        for i in range(matrix.rows):
            for j in range(matrix.columns):
                inputs.append(matrix[i, j])
        return reduce(max_2, inputs)

    if axis == 0:
        result = Matrix(
            1, matrix.columns, signed=matrix.signed, bits=bits)

        for i in range(matrix.columns):
            inputs = []
            for j in range(matrix.rows):
                inputs.append(matrix[j, i])
            result[0, i] = reduce(max_2, inputs)
        return result

    if axis == 1:
        result = Matrix(
            1, matrix.rows, signed=matrix.signed, bits=bits)
        for i in range(matrix.rows):
            inputs = []
            for j in range(matrix.columns):
                inputs.append(matrix[i, j])
            result[0, i] = reduce(max_2, inputs)
        return result

    raise PyrtlError('Axis invalid: expected (None, 0, or 1), got %s' % axis)


def argmax(matrix, axis=None, bits=None):
    ''' Returns the index of the max value of the matrix.

    :param Matrix/Wirevector matrix: the matrix to perform argmax operation on.
        If it is a WireVector, it will return itself
    :param None/int axis: The axis to perform the operation on.
        None refers to argmax of all items. 0 is argmax of the columns. 1 is argmax of rows.
        Defaults to None
    :param int bits: The bits per value of the argmax. Defaults to bits of old matrix
    :return: A WireVector or Matrix representing the argmax value

    NOTE: If there are two indices with the same max value, this function
    picks the first instance.
    '''
    if isinstance(matrix, WireVector):
        return Const(0)

    if not isinstance(matrix, Matrix):
        raise PyrtlError('error: expecting a Matrix or Wirevector for matrix, '
                         'got %s instead' % type(matrix))

    if not isinstance(bits, int) and bits is not None:
        raise PyrtlError('error: expecting an int/None for bits, '
                         'got %s instead' % type(bits))

    if not isinstance(axis, int) and axis is not None:
        raise PyrtlError('error: expecting an int or None for axis, '
                         'got %s instead' % type(axis))

    if bits is None:
        bits = matrix.bits

    if bits <= 0:
        raise PyrtlError('error: bits cannot be negative or zero, '
                         'got %s instead' % bits)

    max_number = max(matrix, axis=axis, bits=bits)
    if axis is None:
        index = Const(0)
        arg = matrix.rows * matrix.columns - 1
        for i in reversed(range(matrix.rows)):
            for j in reversed(range(matrix.columns)):
                index = select(
                    max_number == matrix[i, j], Const(arg), index)
                arg -= 1
        return index
    if axis == 0:
        result = Matrix(
            1, matrix.columns, signed=matrix.signed, bits=bits)
        for i in range(matrix.columns):
            local_max = max_number[0, i]
            index = Const(0)
            arg = matrix.rows - 1
            for j in reversed(range(matrix.rows)):
                index = select(
                    local_max == matrix[j, i], Const(arg), index)
                arg -= 1
            result[0, i] = index
        return result
    if axis == 1:
        result = Matrix(
            1, matrix.rows, signed=matrix.signed, bits=bits)
        for i in range(matrix.rows):
            local_max = max_number[0, i]
            index = Const(0)
            arg = matrix.columns - 1
            for j in reversed(range(matrix.columns)):
                index = select(
                    local_max == matrix[i, j], Const(arg), index)
                arg -= 1
            result[0, i] = index
        return result


def dot(first, second):
    ''' Performs the dot product on two matrices.

    :param Matrix first: the first matrix
    :param Matrix second: the second matrix
    :return: a PyRTL Matrix that contains the dot product of the two PyRTL Matrices

    Specifically, the dot product on two matrices is
        * If either first or second are WireVectors/have both rows and columns
          equal to 1, it is equivalent to multiply
        * If both first and second are both arrays (have rows or columns equal to 1),
          it is inner product of vectors.
        * Otherwise it is the matrix multiplaction between first and second

    NOTE: Row vectors and column vectors are both treated as arrays
    '''
    if not isinstance(first, (WireVector, Matrix)):
        raise PyrtlError('error: expecting a Matrix, '
                         'got %s instead' % type(first))
    if not isinstance(second, (WireVector, Matrix)):
        raise PyrtlError('error: expecting a Matrix/WireVector, '
                         'got %s instead' % type(second))

    # First case when it is multiply
    if isinstance(first, WireVector):
        if isinstance(second, WireVector):
            return first * second
        return second[:, :] * first
    if isinstance(second, WireVector):
        return first[:, :] * second
    if (first.rows == 1 and first.columns == 1) \
            or (second.rows == 1 and second.columns == 1):
        return first[:, :] * second[:, :]

    # Second case when it is Inner Product
    if first.rows == 1:
        if second.rows == 1:
            return sum(first * second)
        if second.columns == 1:
            return sum(first * second.transpose())
    elif first.columns == 1:
        if second.rows == 1:
            return sum(first * second.transpose())
        if second.columns == 1:
            return sum(first * second)

    # Third case when it is Matrix Multiply
    return first.__matmul__(second)


def hstack(*matrices):
    """ Stack matrices in sequence horizontally (column-wise).

    :param list[Matrix] matrices: a list of matrices to concatenate one after another horizontally
    :return Matrix: a new Matrix, with the same number of rows as the original, with
        a bitwidth equal to the max of the bitwidths of all the matrices

    All the matrices must have the same number of rows and same 'signed' value.

    For example::

        m1 = Matrix(2, 3, bits=5,  value=[[1,2,3],
                                          [4,5,6]])
        m2 = Matrix(2, 1, bits=10, value=[[17],
                                          [23]]])
        m3 = hstack(m1, m2)

    m3 looks like::

        [[1,2,3,17],
         [4,5,6,23]]
    """
    if len(matrices) == 0:
        raise PyrtlError("Must supply at least one matrix to hstack()")

    if any([not isinstance(matrix, Matrix) for matrix in matrices]):
        raise PyrtlError("All arguments to hstack must be matrices.")

    if len(matrices) == 1:
        return matrices[0].copy()

    new_rows = matrices[0].rows
    if any([m.rows != new_rows for m in matrices]):
        raise PyrtlError("All matrices being hstacked together must have the same number of rows")

    new_signed = matrices[0].signed
    if any([m.signed != new_signed for m in matrices]):
        raise PyrtlError("All matrices being hstacked together must have the same signedness")

    new_cols = builtins.sum(m.columns for m in matrices)
    new_bits = builtins.max(m.bits for m in matrices)
    new_max_bits = builtins.max(m.max_bits for m in matrices)
    new = Matrix(new_rows, new_cols, new_bits, max_bits=new_max_bits)

    new_c = 0
    for matrix in matrices:
        for c in range(matrix.columns):
            for r in range(matrix.rows):
                new[r, new_c] = matrix[r, c]
            new_c += 1

    return new


def vstack(*matrices):
    """ Stack matrices in sequence vertically (row-wise).

    :param list[Matrix] matrices: a list of matrices to concatenate one after another vertically
    :return Matrix: a new Matrix, with the same number of columns as the original, with
        a bitwidth equal to the max of the bitwidths of all the matrices

    All the matrices must have the same number of columns and same 'signed' value.

    For example::

        m1 = Matrix(2, 3, bits=5,  value=[[1,2,3],
                                          [4,5,6]])
        m2 = Matrix(1, 3, bits=10, value=[[7,8,9]])
        m3 = vstack(m1, m2)

    m3 looks like::

        [[1,2,3],
         [4,5,6],
         [7,8,9]]
    """
    if len(matrices) == 0:
        raise PyrtlError("Must supply at least one matrix to hstack()")

    if any([not isinstance(matrix, Matrix) for matrix in matrices]):
        raise PyrtlError("All arguments to vstack must be matrices.")

    if len(matrices) == 1:
        return matrices[0].copy()

    new_cols = matrices[0].columns
    if any([m.columns != new_cols for m in matrices]):
        raise PyrtlError("All matrices being vstacked together must have the "
                         "same number of columns")

    new_signed = matrices[0].signed
    if any([m.signed != new_signed for m in matrices]):
        raise PyrtlError("All matrices being hstacked together must have the same signedness")

    new_rows = builtins.sum(m.rows for m in matrices)
    new_bits = builtins.max(m.bits for m in matrices)
    new_max_bits = builtins.max(m.max_bits for m in matrices)
    new = Matrix(new_rows, new_cols, new_bits, max_bits=new_max_bits)

    new_r = 0
    for matrix in matrices:
        for r in range(matrix.rows):
            for c in range(matrix.columns):
                new[new_r, c] = matrix[r, c]
            new_r += 1

    return new


def concatenate(matrices, axis=0):
    """ Join a sequence of matrices along an existing axis.

    :param list[Matrix] matrices: a list of matrices to concatenate one after another
    :param int axix: axis along which to join; 0 is horizontally, 1 is vertically (defaults to 0)
    :return: a new Matrix composed of the given matrices joined together

    This function essentially wraps hstack/vstack.
    """
    if axis == 0:
        return hstack(*matrices)
    elif axis == 1:
        return vstack(*matrices)
    else:
        raise PyrtlError("Only allowable axes are 0 or 1")


def matrix_wv_to_list(matrix_wv, rows, columns, bits):
    ''' Convert a wirevector representing a matrix into a Python list of lists.

    :param WireVector matrix_wv: result of calling to_wirevector() on a Matrix object
    :param int rows: number of rows in the matrix matrix_wv represents
    :param int columns: number of columns in the matrix matrix_wv represents
    :param int bits: number of bits in each element of the matrix matrix_wv represents
    :return list[list[int]]: a Python list of lists

    This is useful when printing the value of a wire you've inspected
    during Simulation that you know represnts a matrix.

    Example::

        values = [[1, 2, 3], [4, 5, 6]]
        rows = 2
        cols = 3
        bits = 4
        m = Matrix.Matrix(rows, cols, bits, values=values)

        output = Output(name='output')
        output <<= m.to_wirevector()

        sim = Simulation()
        sim.step({})

        raw_matrix = Matrix.matrix_wv_to_list(sim.inspect('output'), rows, cols, bits)
        print(raw_matrix)

        # Produces:
        # [[1, 2, 3], [4, 5, 6]]
    '''
    value = bin(matrix_wv)[2:].zfill(rows * columns * bits)

    result = [[0 for _ in range(columns)]
              for _ in range(rows)]

    bit_pointer = 0
    for i in range(rows):
        for j in range(columns):
            int_value = int(value[bit_pointer: bit_pointer + bits], 2)
            result[i][j] = int_value
            bit_pointer += bits
    return result


def list_to_int(matrix, n_bits):
    ''' Convert a Python matrix (a list of lists) into an integer.

    :param list[list[int]] matrix: a pure Python list of lists representing a matrix
    :param int n_bits: number of bits to be used to represent each element; if an
        element doesn't fit in n_bits, it truncates the most significant bits
    :return int: a N*n_bits wide wirevector containing the elements of `matrix`,
        where N is the number of elements in `matrix`

    Integers that are signed will automatically be converted to their two's complement form.

    This function is helpful for turning a pure Python list of lists
    into a integer suitable for creating a Constant wirevector that can
    be passed in to as a Matrix intializer's `value` argument, or for
    passing into a Simulation's step function for a particular input wire.

    For example, calling Matrix.list_to_int([3, 5], [7, 9], 4) produces 13,689,
    which in binary looks like this::

        0011 0101 0111 1001

    Note how the elements of the list of lists were added, 4 bits at a time,
    in row order, such that the element at row 0, column 0 is in the most significant
    4 bits, and the element at row 1, column 1 is in the least significant 4 bits.

    Here's an example of using it in simulation::

        a_vals = [[0, 1], [2, 3]]
        b_vals = [[2, 4, 6], [8, 10, 12]]

        a_in = pyrtl.Input(4 * 4, 'a_in')
        b_in = pyrtl.Input(6 * 4, 'b_in')
        a = Matrix.Matrix(2, 2, 4, value=a_in)
        b = Matrix.Matrix(2, 3, 4, value=b_in)
        ...

        sim = pyrtl.Simulation()
        sim.step({
            'a_in': Matrix.list_to_int(a_vals)
            'b_in': Matrix.list_to_int(b_vals)
        })
    '''
    if n_bits <= 0:
        raise PyrtlError("Number of bits per element must be positive, instead got %d" % n_bits)

    result = 0

    for i in range(len(matrix)):
        for j in range(len(matrix[0])):
            val = formatted_str_to_val(str(matrix[i][j]), 's' + str(n_bits))
            result = (result << n_bits) | val

    return result



import pyrtl


def prioritized_mux(selects, vals):
    """ Returns the value in the first wire for which its select bit is 1

    :param [WireVector] selects: a list of WireVectors signaling whether
        a wire should be chosen
    :param [WireVector] vals: values to return when the corresponding select
        value is 1
    :return: WireVector

    If none of the items are high, the last val is returned
    """
    if len(selects) != len(vals):
        raise pyrtl.PyrtlError("Number of select and val signals must match")
    if len(vals) == 0:
        raise pyrtl.PyrtlError("Must have a signal to mux")
    if len(vals) == 1:
        return vals[0]
    else:
        half = len(vals) // 2
        return pyrtl.select(pyrtl.rtl_any(*selects[:half]),
                            truecase=prioritized_mux(selects[:half], vals[:half]),
                            falsecase=prioritized_mux(selects[half:], vals[half:]))


def _is_equivalent(w1, w2):
    if isinstance(w1, pyrtl.Const) & isinstance(w2, pyrtl.Const):
        return (w1.val == w2.val) & (w1.bitwidth == w2.bitwidth)
    return w1 is w2


SparseDefault = "default"


def sparse_mux(sel, vals):
    """ Mux that avoids instantiating unnecessary mux_2s when possible.

    :param WireVector sel: Select wire, determines what is selected on a given cycle
    :param dictionary vals: dictionary of values at mux inputs (of type `{int:WireVector}`)
    :return: WireVector that signifies the change

    This mux supports not having a full specification. Indices that are not
    specified are treated as don't-cares

    It also supports a specified default value, SparseDefault
    """
    import numbers

    max_val = 2**len(sel) - 1
    if SparseDefault in vals:
        default_val = vals[SparseDefault]
        del vals[SparseDefault]
        for i in range(max_val + 1):
            if i not in vals:
                vals[i] = default_val

    for key in vals.keys():
        if not isinstance(key, numbers.Integral):
            raise pyrtl.PyrtlError("value %s nust be either an integer or 'default'" % str(key))
        if key < 0 or key > max_val:
            raise pyrtl.PyrtlError("value %s is out of range of the sel wire" % str(key))

    return _sparse_mux(sel, vals)


def _sparse_mux(sel, vals):
    """ Mux that avoids instantiating unnecessary mux_2s when possible.

    :param WireVector sel: Select wire, determines what is selected on a given cycle
    :param {int: WireVector} vals: dictionary to store the values that are
    :return: Wirevector that signifies the change

    This mux supports not having a full specification. indices that are not
    specified are treated as Don't Cares
    """
    items = list(vals.values())
    if len(vals) <= 1:
        if len(vals) == 0:
            raise pyrtl.PyrtlError("Needs at least one parameter for val")
        return items[0]

    if len(sel) == 1:
        try:
            false_result = vals[0]
            true_result = vals[1]
        except KeyError:
            raise pyrtl.PyrtlError("Failed to retrieve values for smartmux. "
                                   "The length of sel might be wrong")
    else:
        half = 2**(len(sel) - 1)

        first_dict = {indx: wire for indx, wire in vals.items() if indx < half}
        second_dict = {indx - half: wire for indx, wire in vals.items() if indx >= half}
        if not len(first_dict):
            return sparse_mux(sel[:-1], second_dict)
        if not len(second_dict):
            return sparse_mux(sel[:-1], first_dict)

        false_result = sparse_mux(sel[:-1], first_dict)
        true_result = sparse_mux(sel[:-1], second_dict)
    if _is_equivalent(false_result, true_result):
        return true_result
    return pyrtl.select(sel[-1], falsecase=false_result, truecase=true_result)


class MultiSelector(object):
    """ The MultiSelector allows you to specify multiple wire value results
    for a single select wire.

    Useful for processors, finite state machines and other places where the
    result of many wire values are determined by a common wire signal
    (such as a 'state' wire).

    Example::

        with muxes.MultiSelector(select, res0, res1, res2, ...) as ms:
            ms.option(val1, data0, data1, data2, ...)
            ms.option(val2, data0_2, data1_2, data2_2, ...)

    This means that when the select wire equals the val1 wire
    the results will have the values in the coresponding data wires
    (all ints are converted to wires)
    """
    def __init__(self, signal_wire, *dest_wires):
        self._final = False
        self.dest_wires = dest_wires
        self.signal_wire = signal_wire
        self.instructions = []
        self.dest_instrs_info = {dest_w: [] for dest_w in dest_wires}

    def __enter__(self):
        """ For compatibility with `with` statements, which is the recommended
        method of using a MultiSelector.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.finalize()
        else:
            print("The MultiSelector was not finalized due to uncaught exception")

    def _check_finalized(self):
        if self._final:
            raise pyrtl.PyrtlError("Cannot change InstrConnector, already finalized")

    def option(self, select_val, *data_signals):
        self._check_finalized()
        instr, ib = pyrtl.infer_val_and_bitwidth(select_val, self.signal_wire.bitwidth)
        if instr in self.instructions:
            raise pyrtl.PyrtlError("instruction %s already exists" % str(select_val))
        self.instructions.append(instr)
        self._add_signal(data_signals)

    def default(self, *data_signals):
        self._check_finalized()
        self.instructions.append(SparseDefault)
        self._add_signal(data_signals)

    def _add_signal(self, data_signals):
        self._check_finalized()
        if len(data_signals) != len(self.dest_wires):
            raise pyrtl.PyrtlError("Incorrect number of data_signals for "
                                   "instruction received {} , expected {}"
                                   .format(len(data_signals), len(self.dest_wires)))

        for dw, sig in zip(self.dest_wires, data_signals):
            data_signal = pyrtl.as_wires(sig, dw.bitwidth)
            self.dest_instrs_info[dw].append(data_signal)

    def finalize(self):
        """ Connects the wires.
        """
        self._check_finalized()
        self._final = True

        for dest_w, values in self.dest_instrs_info.items():
            mux_vals = dict(zip(self.instructions, values))
            dest_w <<= sparse_mux(self.signal_wire, mux_vals)


def demux(select):
    """ Demultiplexes a wire of arbitrary bitwidth

    :param WireVector select: indicates which wire to set on
    :return (WireVector, ...): a tuple of wires corresponding to each demultiplexed wire
    """
    if len(select) == 1:
        return _demux_2(select)

    wires = demux(select[:-1])
    sel = select[-1]
    not_select = ~sel
    zero_wires = tuple(not_select & w for w in wires)
    one_wires = tuple(sel & w for w in wires)
    return zero_wires + one_wires


def _demux_2(select):
    assert len(select) == 1
    return ~select, select




# error types thrown
from .pyrtlexceptions import PyrtlError
from .pyrtlexceptions import PyrtlInternalError

# core rtl constructs
from .core import LogicNet
from .core import Block
from .core import PostSynthBlock
from .core import working_block
from .core import reset_working_block
from .core import set_working_block
from .core import temp_working_block
from .core import set_debug_mode

# convenience classes for building hardware
from .wire import WireVector
from .wire import Input, Output
from .wire import Const
from .wire import Register

# helper functions
from .helperfuncs import input_list
from .helperfuncs import output_list
from .helperfuncs import register_list
from .helperfuncs import wirevector_list
from .helperfuncs import log2
from .helperfuncs import truncate
from .helperfuncs import match_bitpattern
from .helperfuncs import bitpattern_to_val
from .helperfuncs import chop
from .helperfuncs import val_to_signed_integer
from .helperfuncs import val_to_formatted_str
from .helperfuncs import formatted_str_to_val
from .helperfuncs import infer_val_and_bitwidth
from .helperfuncs import probe
from .helperfuncs import rtl_assert
from .helperfuncs import check_rtl_assertions
from .helperfuncs import find_loop
from .helperfuncs import find_and_print_loop

from .corecircuits import and_all_bits
from .corecircuits import or_all_bits
from .corecircuits import xor_all_bits
from .corecircuits import rtl_any
from .corecircuits import rtl_all
from .corecircuits import mux
from .corecircuits import select
from .corecircuits import concat
from .corecircuits import concat_list
from .corecircuits import parity
from .corecircuits import tree_reduce
from .corecircuits import as_wires
from .corecircuits import match_bitwidth
from .corecircuits import enum_mux
from .corecircuits import bitfield_update
from .corecircuits import bitfield_update_set
from .corecircuits import signed_add
from .corecircuits import signed_mult
from .corecircuits import signed_lt
from .corecircuits import signed_le
from .corecircuits import signed_gt
from .corecircuits import signed_ge
from .corecircuits import shift_left_arithmetic
from .corecircuits import shift_right_arithmetic
from .corecircuits import shift_left_logical
from .corecircuits import shift_right_logical

# memory blocks
from .memory import MemBlock
from .memory import RomBlock

# conditional updates
from .conditional import conditional_assignment
from .conditional import otherwise
from .conditional import currently_under_condition

# block simulation support
from .simulation import Simulation
from .simulation import FastSimulation
from .simulation import SimulationTrace
from .compilesim import CompiledSimulation

# block visualization output formats
from .visualization import output_to_trivialgraph
from .visualization import graphviz_detailed_namer
from .visualization import output_to_graphviz
from .visualization import output_to_svg
from .visualization import block_to_graphviz_string
from .visualization import block_to_svg
from .visualization import trace_to_html
from .visualization import net_graph

# import from and export to file format routines
from .importexport import input_from_verilog
from .importexport import output_to_verilog
from .importexport import OutputToVerilog
from .importexport import output_verilog_testbench
from .importexport import input_from_blif
from .importexport import output_to_firrtl
from .importexport import input_from_iscas_bench

# different transform passes
from .passes import common_subexp_elimination
from .passes import constant_propagation
from .passes import synthesize
from .passes import nand_synth
from .passes import and_inverter_synth
from .passes import optimize
from .passes import one_bit_selects
from .passes import two_way_concat
from .passes import direct_connect_outputs
from .passes import two_way_fanout

from .transform import net_transform
from .transform import wire_transform
from .transform import copy_block
from .transform import clone_wire
from .transform import replace_wires
from .transform import replace_wire_fast

# analysis and estimation functions
from .analysis import area_estimation
from .analysis import TimingAnalysis
from .analysis import yosys_area_delay
from .analysis import paths
from .analysis import distance
from .analysis import fanout



from __future__ import absolute_import

import pyrtl


def match_bitwidth(*args):
    # TODO: allow for custom bit extension functions
    """ Matches the bitwidth of all of the input arguments.

    :param WireVector args: input arguments
    :return: tuple of `args` in order with extended bits
    """
    return pyrtl.match_bitwidth(*args)


def partition_wire(wire, partition_size):
    """ Partitions a wire into a list of N wires of size 'partition_size'.

    :param wire: Wire to partition
    :param partition_size: Integer representing size of each partition

    The wire's bitwidth must be evenly divisible by 'parition_size'.
    """
    if len(wire) % partition_size != 0:
        raise pyrtl.PyrtlError("Wire {} cannot be evenly partitioned into items of size {}"
                               .format(wire, partition_size))
    return [wire[offset:offset + partition_size] for offset in range(0, len(wire), partition_size)]


def str_to_int_array(string, base=16):
    """
    Converts a string to an array of integer values according to the
    base specified (int numbers must be whitespace delimited).\n
    Example: "13 a3 3c" => [0x13, 0xa3, 0x3c]

    :return: [int]
    """

    int_strings = string.split()
    return [int(int_str, base) for int_str in int_strings]


def twos_comp_repr(val, bitwidth):
    """
    Converts a value to its two's-complement (positive) integer representation using a
    given bitwidth (only converts the value if it is negative).

    :param val: Integer literal to convert to two's complement
    :param bitwidth: Size of val in bits

    For use with Simulation.step() etc. in passing negative numbers, which it does not accept.
    """
    correctbw = abs(val).bit_length() + 1
    if bitwidth < correctbw:
        raise pyrtl.PyrtlError("please choose a larger target bitwidth")
    if val >= 0:
        return val
    else:
        return (~abs(val) & (2 ** bitwidth - 1)) + 1  # flip the bits and add one


def rev_twos_comp_repr(val, bitwidth):
    """
    Takes a two's-complement represented value and
    converts it to a signed integer based on the provided bitwidth.
    For use with Simulation.inspect() etc. when expecting negative numbers,
    which it does not recognize
    """
    valbl = val.bit_length()
    if bitwidth < val.bit_length() or val == 2 ** (bitwidth - 1):
        raise pyrtl.PyrtlError("please choose a larger target bitwidth")
    if bitwidth == valbl:  # MSB is a 1, value is negative
        return -((~val & (2 ** bitwidth - 1)) + 1)  # flip the bits, add one, and make negative
    else:
        return val


def _shifted_reg_next(reg, direct, num=1):
    """
    Creates a shifted 'next' property for shifted (left or right) register.\n
    Use: `myReg.next = shifted_reg_next(myReg, 'l', 4)`

    :param string direct: direction of shift, either 'l' or 'r'
    :param int num: number of shifts
    :return: Register containing reg's (shifted) next state
    """
    if direct == 'l':
        if num >= len(reg):
            return 0
        else:
            return pyrtl.concat(reg, pyrtl.Const(0, num))
    elif direct == 'r':
        if num >= len(reg):
            return 0
        else:
            return reg[num:]
    else:
        raise pyrtl.PyrtlError("direction must be specified with 'direct'"
                               "parameter as either 'l' or 'r'")



from __future__ import absolute_import
import pyrtl
from . import libutils


def kogge_stone(a, b, cin=0):
    """
    Creates a Kogge-Stone adder given two inputs

    :param WireVector a, b: The two WireVectors to add up (bitwidths don't need to match)
    :param cin: An optimal carry in WireVector or value
    :return: a Wirevector representing the output of the adder

    The Kogge-Stone adder is a fast tree-based adder with O(log(n))
    propagation delay, useful for performance critical designs. However,
    it has O(n log(n)) area usage, and large fan out.
    """
    a, b = libutils.match_bitwidth(a, b)

    prop_orig = a ^ b
    prop_bits = [i for i in prop_orig]
    gen_bits = [i for i in a & b]
    prop_dist = 1

    # creation of the carry calculation
    while prop_dist < len(a):
        for i in reversed(range(prop_dist, len(a))):
            prop_old = prop_bits[i]
            gen_bits[i] = gen_bits[i] | (prop_old & gen_bits[i - prop_dist])
            if i >= prop_dist * 2:  # to prevent creating unnecessary nets and wires
                prop_bits[i] = prop_old & prop_bits[i - prop_dist]
        prop_dist *= 2

    # assembling the result of the addition
    # preparing the cin (and conveniently shifting the gen bits)
    gen_bits.insert(0, pyrtl.as_wires(cin))
    return pyrtl.concat_list(gen_bits) ^ prop_orig


def one_bit_add(a, b, cin=0):
    return pyrtl.concat(*_one_bit_add_no_concat(a, b, cin))


def _one_bit_add_no_concat(a, b, cin=0):
    cin = pyrtl.as_wires(cin)  # to make sure that an int cin doesn't break things
    assert len(a) == len(b) == len(cin) == 1
    sum = a ^ b ^ cin
    cout = a & b | a & cin | b & cin
    return cout, sum


def half_adder(a, b):
    assert len(a) == len(b) == 1
    sum = a ^ b
    cout = a & b
    return cout, sum


def ripple_add(a, b, cin=0):
    if len(a) < len(b):  # make sure that b is the shorter wire
        b, a = a, b
    cin = pyrtl.as_wires(cin)
    if len(a) == 1:
        return one_bit_add(a, b, cin)
    else:
        ripplecarry = one_bit_add(a[0], b[0], cin)
        if len(b) == 1:
            msbits = ripple_half_add(a[1:], ripplecarry[1])
        else:
            msbits = ripple_add(a[1:], b[1:], ripplecarry[1])
        return pyrtl.concat(msbits, ripplecarry[0])


def ripple_half_add(a, cin=0):
    cin = pyrtl.as_wires(cin)
    if len(a) == 1:
        return pyrtl.concat(*half_adder(a, cin))
    else:
        ripplecarry = half_adder(a[0], cin)
        msbits = ripple_half_add(a[1:], ripplecarry[0])
        return pyrtl.concat(msbits, ripplecarry[1])


def carrysave_adder(a, b, c, final_adder=ripple_add):
    """
    Adds three wirevectors up in an efficient manner
    :param WireVector a, b, c : the three wires to add up
    :param function final_adder : The adder to use to do the final addition
    :return: a wirevector with length 2 longer than the largest input
    """
    a, b, c = libutils.match_bitwidth(a, b, c)
    partial_sum = a ^ b ^ c
    shift_carry = (a | b) & (a | c) & (b | c)
    return pyrtl.concat(final_adder(partial_sum[1:], shift_carry), partial_sum[0])


def cla_adder(a, b, cin=0, la_unit_len=4):
    """
    Carry Lookahead Adder
    :param int la_unit_len: the length of input that every unit processes

    A Carry LookAhead Adder is an adder that is faster than
    a ripple carry adder, as it calculates the carry bits faster.
    It is not as fast as a Kogge-Stone adder, but uses less area.
    """
    a, b = pyrtl.match_bitwidth(a, b)
    if len(a) <= la_unit_len:
        sum, cout = _cla_adder_unit(a, b, cin)
        return pyrtl.concat(cout, sum)
    else:
        sum, cout = _cla_adder_unit(a[0:la_unit_len], b[0:la_unit_len], cin)
        msbits = cla_adder(a[la_unit_len:], b[la_unit_len:], cout, la_unit_len)
        return pyrtl.concat(msbits, sum)


def _cla_adder_unit(a, b, cin):
    """
    Carry generation and propogation signals will be calculated only using
    the inputs; their values don't rely on the sum.  Every unit generates
    a cout signal which is used as cin for the next unit.
    """
    gen = a & b
    prop = a ^ b
    assert len(prop) == len(gen)

    carry = [gen[0] | prop[0] & cin]
    sum_bit = prop[0] ^ cin

    cur_gen = gen[0]
    cur_prop = prop[0]
    for i in range(1, len(prop)):
        cur_gen = gen[i] | (prop[i] & cur_gen)
        cur_prop = cur_prop & prop[i]
        sum_bit = pyrtl.concat(prop[i] ^ carry[i - 1], sum_bit)
        carry.append(gen[i] | (prop[i] & carry[i - 1]))
    cout = cur_gen | (cur_prop & cin)
    return sum_bit, cout


def wallace_reducer(wire_array_2, result_bitwidth, final_adder=kogge_stone):
    """
    The reduction and final adding part of a dada tree. Useful for adding many numbers together
    The use of single bitwidth wires is to allow for additional flexibility

    :param [[Wirevector]] wire_array_2: An array of arrays of single bitwidth
        wirevectors
    :param int result_bitwidth: The bitwidth you want for the resulting wire.
        Used to eliminate unnessary wires.
    :param final_adder: The adder used for the final addition
    :return: wirevector of length result_wirevector
    """
    # verification that the wires are actually wirevectors of length 1
    for wire_set in wire_array_2:
        for a_wire in wire_set:
            if not isinstance(a_wire, pyrtl.WireVector) or len(a_wire) != 1:
                raise pyrtl.PyrtlError(
                    "The item {} is not a valid element for the wire_array_2. "
                    "It must be a WireVector of bitwidth 1".format(a_wire))

    while not all(len(i) <= 2 for i in wire_array_2):
        deferred = [[] for weight in range(result_bitwidth + 1)]
        for i, w_array in enumerate(wire_array_2):  # Start with low weights and start reducing
            while len(w_array) >= 3:
                cout, sum = _one_bit_add_no_concat(*(w_array.pop(0) for j in range(3)))
                deferred[i].append(sum)
                deferred[i + 1].append(cout)

            if len(w_array) == 2:
                cout, sum = half_adder(*w_array)
                deferred[i].append(sum)
                deferred[i + 1].append(cout)
            else:
                deferred[i].extend(w_array)
        wire_array_2 = deferred[:result_bitwidth]

    # At this stage in the multiplication we have only 2 wire vectors left.
    # now we need to add them up
    result = _sparse_adder(wire_array_2, final_adder)
    if len(result) > result_bitwidth:
        return result[:result_bitwidth]
    else:
        return result


def dada_reducer(wire_array_2, result_bitwidth, final_adder=kogge_stone):
    """
    The reduction and final adding part of a dada tree. Useful for adding many numbers together
    The use of single bitwidth wires is to allow for additional flexibility

    :param [[Wirevector]] wire_array_2: An array of arrays of single bitwidth
        wirevectors
    :param int result_bitwidth: The bitwidth you want for the resulting wire.
        Used to eliminate unnessary wires.
    :param final_adder: The adder used for the final addition
    :return: wirevector of length result_wirevector
    """
    import math
    # verification that the wires are actually wirevectors of length 1
    for wire_set in wire_array_2:
        for a_wire in wire_set:
            if not isinstance(a_wire, pyrtl.WireVector) or len(a_wire) != 1:
                raise pyrtl.PyrtlError(
                    "The item {} is not a valid element for the wire_array_2. "
                    "It must be a WireVector of bitwidth 1".format(a_wire))

    max_width = max(len(i) for i in wire_array_2)
    reduction_schedule = [2]
    while reduction_schedule[-1] <= max_width:
        reduction_schedule.append(int(reduction_schedule[-1] * 3 / 2))

    for reduction_target in reversed(reduction_schedule[:-1]):
        deferred = [[] for weight in range(result_bitwidth + 1)]
        last_round = (max(len(i) for i in wire_array_2) == 3)
        for i, w_array in enumerate(wire_array_2):  # Start with low weights and start reducing
            while len(w_array) + len(deferred[i]) > reduction_target:
                if len(w_array) + len(deferred[i]) - reduction_target >= 2:
                    cout, sum = _one_bit_add_no_concat(*(w_array.pop(0) for j in range(3)))
                    deferred[i].append(sum)
                    deferred[i + 1].append(cout)
                else:
                    # if (last_round and len(deferred[i]) % 3 == 1) or (len(deferred[i]) % 3 == 2):
                    # if not(last_round and len(wire_array_2[i + 1]) < 3):
                    cout, sum = half_adder(*(w_array.pop(0) for j in range(2)))
                    deferred[i].append(sum)
                    deferred[i + 1].append(cout)
            deferred[i].extend(w_array)
            if len(deferred[i]) > reduction_target:
                raise pyrtl.PyrtlError("Expected that the code would be able to reduce more wires")
        wire_array_2 = deferred[:result_bitwidth]

    # At this stage in the multiplication we have only 2 wire vectors left.
    # now we need to add them up
    result = _sparse_adder(wire_array_2, final_adder)
    if len(result) > result_bitwidth:
        return result[:result_bitwidth]
    else:
        return result


def _sparse_adder(wire_array_2, adder):
    result = []
    for single_w_index in range(len(wire_array_2)):
        if len(wire_array_2[single_w_index]) == 2:  # Check if the two wire vectors overlap yet
            break
        result.append(wire_array_2[single_w_index][0])

    import six
    wires_to_zip = wire_array_2[single_w_index:]
    add_wires = tuple(six.moves.zip_longest(*wires_to_zip, fillvalue=pyrtl.Const(0)))
    adder_result = adder(pyrtl.concat_list(add_wires[0]), pyrtl.concat_list(add_wires[1]))
    return pyrtl.concat(adder_result, *reversed(result))


"""
Some adders that utilize these tree reducers
"""


def fast_group_adder(wires_to_add, reducer=wallace_reducer, final_adder=kogge_stone):
    """
    A generalization of the carry save adder, this is designed to add many numbers
    together in a both area and time efficient manner. Uses a tree reducer
    to achieve this performance


    :param [WireVector] wires_to_add: an array of wirevectors to add
    :param reducer: the tree reducer to use
    :param final_adder: The two value adder to use at the end
    :return: a wirevector with the result of the addition
      The length of the result is:
      max(len(w) for w in wires_to_add) + ceil(len(wires_to_add))
    """
    import math
    longest_wire_len = max(len(w) for w in wires_to_add)
    result_bitwidth = longest_wire_len + int(math.ceil(math.log(len(wires_to_add), 2)))

    bits = [[] for i in range(longest_wire_len)]

    for wire in wires_to_add:
        for bit_loc, bit in enumerate(wire):
            bits[bit_loc].append(bit)

    return reducer(bits, result_bitwidth, final_adder)



from __future__ import absolute_import
import pyrtl
import math


def barrel_shifter(bits_to_shift, bit_in, direction, shift_dist, wrap_around=0):
    """ Create a barrel shifter that operates on data based on the wire width.

    :param bits_to_shift: the input wire
    :param bit_in: the 1-bit wire giving the value to shift in
    :param direction: a one bit WireVector representing shift direction
        (0 = shift down, 1 = shift up)
    :param shift_dist: WireVector representing offset to shift
    :param wrap_around: ****currently not implemented****
    :return: shifted WireVector
    """
    from pyrtl import concat, select  # just for readability

    if wrap_around != 0:
        raise NotImplementedError

    # Implement with logN stages pyrtl.muxing between shifted and un-shifted values
    final_width = len(bits_to_shift)
    val = bits_to_shift
    append_val = bit_in

    for i in range(len(shift_dist)):
        shift_amt = pow(2, i)  # stages shift 1,2,4,8,...
        if shift_amt < final_width:
            newval = select(direction,
                            concat(val[:-shift_amt], append_val),  # shift up
                            concat(append_val, val[shift_amt:]))  # shift down
            val = select(shift_dist[i],
                         truecase=newval,  # if bit of shift is 1, do the shift
                         falsecase=val)  # otherwise, don't
            # the value to append grows exponentially, but is capped at full width
            append_val = concat(append_val, append_val)[:final_width]
        else:
            # if we are shifting this much, all the data is gone
            val = select(shift_dist[i],
                         truecase=append_val,  # if bit of shift is 1, do the shift
                         falsecase=val)  # otherwise, don't

    return val



import random
import unittest
import six
import os
import sys

import pyrtl
import pyrtl.corecircuits
import pyrtl.helperfuncs
from pyrtl.rtllib import testingutils as utils


# ---------------------------------------------------------------

class TestWireVectorList(unittest.TestCase):
    def setUp(self):
        pass

    def test_input_list_type(self):
        inputs = pyrtl.helperfuncs.input_list('one, two, three')
        self.assertTrue(all(isinstance(inp, pyrtl.Input) for inp in inputs))

    def test_output_list_type(self):
        outputs = pyrtl.helperfuncs.output_list('one, two, three')
        self.assertTrue(all(isinstance(outp, pyrtl.Output) for outp in outputs))

    def test_register_list_type(self):
        registers = pyrtl.helperfuncs.register_list('one, two, three')
        self.assertTrue(all(isinstance(reg, pyrtl.Register) for reg in registers))

    def test_wirevector_list_type(self):
        # Single string of names
        wirevectors = pyrtl.helperfuncs.wirevector_list('one, two, three')
        self.assertTrue(all(isinstance(wire, pyrtl.WireVector) for wire in wirevectors))
        self.assertListEqual([wire.bitwidth for wire in wirevectors], [1, 1, 1])

        # List of names
        wirevectors = pyrtl.helperfuncs.wirevector_list('one, two, three')
        self.assertTrue(all(isinstance(wire, pyrtl.WireVector) for wire in wirevectors))
        self.assertListEqual([wire.bitwidth for wire in wirevectors], [1, 1, 1])

    def test_wirevector_list_bitwidth(self):
        wirevectors = pyrtl.helperfuncs.wirevector_list('one, two, three')
        self.assertListEqual([wire.bitwidth for wire in wirevectors], [1, 1, 1])

        wirevectors = pyrtl.helperfuncs.wirevector_list('one, two, three', 8)
        self.assertListEqual([wire.bitwidth for wire in wirevectors], [8, 8, 8])

    def test_wirevector_list_per_wire_width(self):
        wirevectors = pyrtl.helperfuncs.wirevector_list('one/2, two/4, three/8')
        self.assertListEqual([wire.bitwidth for wire in wirevectors], [2, 4, 8])

        wirevectors = pyrtl.helperfuncs.wirevector_list(['one', 'two', 'three'], [2, 4, 8])
        self.assertListEqual([wire.bitwidth for wire in wirevectors], [2, 4, 8])

    def test_wirevector_list_raise_errors(self):

        with self.assertRaises(ValueError):
            pyrtl.helperfuncs.wirevector_list(['one', 'two', 'three'], [2, 4])

        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.helperfuncs.wirevector_list('one/2, two/4, three/8', 16)

        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.helperfuncs.wirevector_list(['one/2', 'two/4', 'three/8'], [8, 4, 2])


class TestNonCoreHelpers(unittest.TestCase):
    def setUp(self):
        pass

    def test_log2(self):
        self.assertEqual(pyrtl.log2(1), 0)
        self.assertEqual(pyrtl.log2(2), 1)
        self.assertEqual(pyrtl.log2(8), 3)
        self.assertEqual(pyrtl.log2(16), 4)
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.log2(-1)
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.log2(1.5)
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.log2(0)
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.log2(7)
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.log2(9)

    def test_truncate_function(self):
        self.assertEqual(pyrtl.truncate(5, 3), 5)
        self.assertEqual(pyrtl.truncate(9, 3), 1)
        self.assertEqual(pyrtl.truncate(-1, 3), 7)
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.truncate(5, -1)
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.truncate(29, 0)

    def test_val_to_signed_integer(self):
        self.assertEqual(pyrtl.val_to_signed_integer(0b000, 3), 0)
        self.assertEqual(pyrtl.val_to_signed_integer(0b001, 3), 1)
        self.assertEqual(pyrtl.val_to_signed_integer(0b010, 3), 2)
        self.assertEqual(pyrtl.val_to_signed_integer(0b011, 3), 3)
        self.assertEqual(pyrtl.val_to_signed_integer(0b100, 3), -4)
        self.assertEqual(pyrtl.val_to_signed_integer(0b101, 3), -3)
        self.assertEqual(pyrtl.val_to_signed_integer(0b110, 3), -2)
        self.assertEqual(pyrtl.val_to_signed_integer(0b111, 3), -1)

    def test_infer_val_and_bitwidth(self):
        self.assertEqual(pyrtl.infer_val_and_bitwidth(2, bitwidth=5), (2, 5))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(3), (3, 2))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(True), (1, 1))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(False), (0, 1))
        self.assertEqual(pyrtl.infer_val_and_bitwidth("5'd12"), (12, 5))
        self.assertEqual(pyrtl.infer_val_and_bitwidth("5'b10"), (2, 5))
        self.assertEqual(pyrtl.infer_val_and_bitwidth("5'b10"), (2, 5))
        self.assertEqual(pyrtl.infer_val_and_bitwidth("8'B 0110_1100"), (108, 8))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(-3, bitwidth=5), (0b11101, 5))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(3, signed=True), (3, 3))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(-3, signed=True), (5, 3))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(-4, signed=True), (4, 3))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(-3, bitwidth=5, signed=True), (29, 5))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(3, bitwidth=2), (3, 2))

        self.assertEqual(pyrtl.infer_val_and_bitwidth(0), (0, 1))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(1), (1, 1))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(2), (2, 2))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(3), (3, 2))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(4), (4, 3))

        self.assertEqual(pyrtl.infer_val_and_bitwidth(0, signed=True), (0, 1))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(1, signed=True), (1, 2))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(2, signed=True), (2, 3))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(3, signed=True), (3, 3))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(4, signed=True), (4, 4))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(-1, signed=True), (1, 1))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(-2, signed=True), (2, 2))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(-3, signed=True), (5, 3))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(-4, signed=True), (4, 3))
        self.assertEqual(pyrtl.infer_val_and_bitwidth(-5, signed=True), (11, 4))

        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.infer_val_and_bitwidth(-3)
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.infer_val_and_bitwidth(True, signed=True)
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.infer_val_and_bitwidth(3, bitwidth=2, signed=True)


class TestMatchBitpattern(unittest.TestCase):
    def setUp(self):
        random.seed(8492049)
        pyrtl.reset_working_block()

    def check_trace(self, correct_string):
        sim_trace = pyrtl.SimulationTrace()
        sim = pyrtl.Simulation(tracer=sim_trace)
        for i in range(8):
            sim.step({})
        output = six.StringIO()
        sim_trace.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), correct_string)

    def test_pattern_type_or_length_mismatch(self):
        instr = pyrtl.WireVector(name='instr', bitwidth=8)
        with self.assertRaises(pyrtl.PyrtlError):
            o, _ = pyrtl.match_bitpattern(instr, '000100010')
        with self.assertRaises(pyrtl.PyrtlError):
            o, _ = pyrtl.match_bitpattern(instr, '0001000')
        with self.assertRaises(pyrtl.PyrtlError):
            o, _ = pyrtl.match_bitpattern(instr, '0b00010001')
        with self.assertRaises(pyrtl.PyrtlError):
            o, _ = pyrtl.match_bitpattern(instr, 0b000100010)
        with self.assertRaises(pyrtl.PyrtlError):
            o, _ = pyrtl.match_bitpattern(instr, '')
        with self.assertRaises(pyrtl.PyrtlError):
            o, _ = pyrtl.match_bitpattern(instr, None)
        with self.assertRaises(pyrtl.PyrtlError):
            o, _ = pyrtl.match_bitpattern(instr, instr)

    def test_match_bitwidth_does_simulation_correct(self):
        r = pyrtl.Register(6, 'r')
        r.next <<= r + 3
        # 000000 -> 000011 -> 000110 -> 001001 -> 001100 -> 001111 -> 010010 -> 010101
        a, matches = pyrtl.match_bitpattern(r, '00_?0 ?1')
        o = pyrtl.Output(name='o')
        o <<= a
        self.assertEqual(matches, ())  # '?' wildcard doesn't bind to a field
        self.check_trace('o 01010000\nr 036912151821\n')

    def test_match_bitwidth_simulates_no_ones_in_pattern(self):
        r = pyrtl.Register(6, 'r')
        r.next <<= r + 3
        # 000000 -> 000011 -> 000110 -> 001001 -> 001100 -> 001111 -> 010010 -> 010101
        a, _ = pyrtl.match_bitpattern(r, '00??00')
        o = pyrtl.Output(name='o')
        o <<= a
        self.check_trace('o 10001000\nr 036912151821\n')

    def test_match_bitwidth_simulates_no_zeroes_in_pattern(self):
        r = pyrtl.Register(6, 'r')
        r.next <<= r + 3
        # 000000 -> 000011 -> 000110 -> 001001 -> 001100 -> 001111 -> 010010 -> 010101
        a, _ = pyrtl.match_bitpattern(r, '?1??1?')
        o = pyrtl.Output(name='o')
        o <<= a
        self.check_trace('o 00000010\nr 036912151821\n')

    def test_match_bitwidth_simulates_only_wildcards(self):
        r = pyrtl.Register(6, 'r')
        r.next <<= r + 3
        # 000000 -> 000011 -> 000110 -> 001001 -> 001100 -> 001111 -> 010010 -> 010101
        a, _ = pyrtl.match_bitpattern(r, '??????')
        o = pyrtl.Output(name='o')
        o <<= a
        self.check_trace('o 11111111\nr 036912151821\n')

    def test_match_bitwidth_with_consecutive_fields(self):
        r = pyrtl.Register(6, 'r')
        r.next <<= r + 3
        # 000000 -> 000011 -> 000110 -> 001001 -> 001100 -> 001111 -> 010010 -> 010101
        m, (b, a) = pyrtl.match_bitpattern(r, 'bb0a?1')
        pyrtl.probe(b, 'b')
        pyrtl.probe(a, 'a')
        o = pyrtl.Output(name='o')
        o <<= m
        self.check_trace(
            'a 00101101\n'
            'b 00000011\n'
            'o 01000001\n'
            'r 036912151821\n'
        )

    def test_match_bitwidth_with_non_consecutive_field(self):
        r = pyrtl.Register(6, 'r')
        r.next <<= r + 3
        # 000000 -> 000011 -> 000110 -> 001001 -> 001100 -> 001111 -> 010010 -> 010101
        m, (b, a) = pyrtl.match_bitpattern(r, 'bb0ab1')
        pyrtl.probe(b, 'b')
        pyrtl.probe(a, 'a')
        o = pyrtl.Output(name='o')
        o <<= m
        self.check_trace(
            'a 00101101\n'
            'b 01100132\n'
            'o 01000001\n'
            'r 036912151821\n'
        )

    def test_match_bitwidth_with_several_non_consecutive_fields(self):
        r = pyrtl.Register(6, 'r')
        r.next <<= r + 3
        # 000000 -> 000011 -> 000110 -> 001001 -> 001100 -> 001111 -> 010010 -> 010101
        m, (a, b) = pyrtl.match_bitpattern(r, 'ab0aba')
        pyrtl.probe(a, 'a')
        pyrtl.probe(b, 'b')
        o = pyrtl.Output(name='o')
        o <<= m
        self.check_trace(
            'a 01212303\n'
            'b 01100132\n'
            'o 11100011\n'
            'r 036912151821\n'
        )

    def test_match_bitwidth_with_all_fields(self):
        r = pyrtl.Register(6, 'r')
        r.next <<= r + 3
        # 000000 -> 000011 -> 000110 -> 001001 -> 001100 -> 001111 -> 010010 -> 010101
        m, (a, b, c, d, e, f) = pyrtl.match_bitpattern(r, 'abcdef')
        pyrtl.probe(a, 'a')
        pyrtl.probe(b, 'b')
        pyrtl.probe(c, 'c')
        pyrtl.probe(d, 'd')
        pyrtl.probe(e, 'e')
        pyrtl.probe(f, 'f')
        o = pyrtl.Output(name='o')
        o <<= m
        self.check_trace(
            'a 00000000\n'
            'b 00000011\n'
            'c 00011100\n'
            'd 00101101\n'
            'e 01100110\n'
            'f 01010101\n'
            'o 11111111\n'
            'r 036912151821\n'
        )

    def check_all_accesses_valid(self):
        sim = pyrtl.Simulation()
        sim.step_multiple({'i': [0b11010, 0b00011, 0b01101]})
        output = six.StringIO()
        sim.tracer.print_trace(output, compact=True)
        self.assertEqual(
            output.getvalue(),
            '  i 26313\n'
            'out 313\n'
        )

    def test_match_bitwidth_with_pattern_matched_fields(self):
        i = pyrtl.Input(5, 'i')
        out = pyrtl.Output(2, 'out')

        with pyrtl.conditional_assignment:
            with pyrtl.match_bitpattern(i, '1a?a0') as (a,):
                out |= a
            with pyrtl.match_bitpattern(i, 'b0?1b') as (b,):
                out |= b
            with pyrtl.match_bitpattern(i, 'ba1ab') as (b, a):
                out |= a + b

        self.check_all_accesses_valid()

    def test_match_bitwidth_with_pattern_matched_fields_by_name(self):
        i = pyrtl.Input(5, 'i')
        out = pyrtl.Output(2, 'out')

        with pyrtl.conditional_assignment:
            with pyrtl.match_bitpattern(i, '1a?a0') as x:
                out |= x.a
            with pyrtl.match_bitpattern(i, 'b0?1b') as x:
                out |= x.b
            with pyrtl.match_bitpattern(i, 'ba1ab') as x:
                out |= x.a + x.b

        self.check_all_accesses_valid()

    def test_match_bitwidth_with_pattern_matched_fields_with_field_map(self):
        i = pyrtl.Input(5, 'i')
        out = pyrtl.Output(2, 'out')

        field_map = {'a': 'field1', 'b': 'field2'}
        with pyrtl.conditional_assignment:
            with pyrtl.match_bitpattern(i, '1a?a0', field_map) as x:
                out |= x.field1
            with pyrtl.match_bitpattern(i, 'b0?1b', field_map) as x:
                out |= x.field2
            with pyrtl.match_bitpattern(i, 'ba1ab', field_map) as x:
                out |= x.field1 + x.field2

        self.check_all_accesses_valid()

    def test_match_bitwidth_with_pattern_matched_fields_with_bad_field_map(self):
        i = pyrtl.Input(5, 'i')
        out = pyrtl.Output(2, 'out')

        field_map = {'a': 'field1', 'c': 'field2'}
        with self.assertRaises(pyrtl.PyrtlError):
            with pyrtl.conditional_assignment:
                with pyrtl.match_bitpattern(i, 'b0?1b', field_map) as x:
                    out |= x.field2


class TestBitpatternToVal(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    @staticmethod
    def r5_br_immed(x):
        # i is the 12th bit, k is the 11th bit, and j is the bottom 10 bits of the immediate field
        return {'i': (x >> 11) & 0x1, 'k': (x >> 10) & 0x1, 'j': x & ((1 << 10) - 1)}

    def test_ordered_fields(self):
        self.assertEqual(
            pyrtl.bitpattern_to_val('0000000sssssrrrrr000ddddd0110011', 1, 2, 3),  # RISC-V ADD
            0b00000000000100010000000110110011
        )

        self.assertEqual(
            pyrtl.bitpattern_to_val('iiiiiiisssssrrrrr010iiiii0100011', 1, 3, 4),  # RISC-V SW
            0b00000000001100100010000010100011
        )

        m = TestBitpatternToVal.r5_br_immed(-5)
        self.assertEqual(
            pyrtl.bitpattern_to_val(
                'ijjjjjjsssssrrrrr100jjjjk1100011', m['i'], m['j'], 2, 3, m['k']
            ),  # RISC-V BLT
            0b11111110001000011100101111100011
        )

    def test_named_fields(self):
        self.assertEqual(
            pyrtl.bitpattern_to_val(
                '0000000sssssrrrrr000ddddd0110011', s=1, r=2, d=3
            ),  # RISC-V ADD
            0b00000000000100010000000110110011
        )

        self.assertEqual(
            pyrtl.bitpattern_to_val('iiiiiiisssssrrrrr010iiiii0100011', i=1, s=3, r=4),  # RISC-V SW
            0b00000000001100100010000010100011
        )

        self.assertEqual(
            pyrtl.bitpattern_to_val(
                'ijjjjjjsssssrrrrr100jjjjk1100011',
                s=2, r=3, **TestBitpatternToVal.r5_br_immed(-5)
            ),  # RISC-V BLT
            0b11111110001000011100101111100011
        )

    def test_named_fields_with_field_map(self):
        field_map = {
            's': 'rs2',
            'r': 'rs1',
            'd': 'rd',
            'i': 'imm',
        }
        self.assertEqual(
            pyrtl.bitpattern_to_val(
                '0000000sssssrrrrr000ddddd0110011', rs2=1, rs1=2, rd=3, field_map=field_map
            ),  # RISC-V ADD
            0b00000000000100010000000110110011
        )

        self.assertEqual(
            pyrtl.bitpattern_to_val(
                'iiiiiiisssssrrrrr010iiiii0100011', imm=1, rs2=3, rs1=4, field_map=field_map
            ),  # RISC-V SW
            0b00000000001100100010000010100011
        )

    def test_fields_all_different(self):
        self.assertEqual(
            pyrtl.bitpattern_to_val('abcdefg', a=1, b=0, c=1, d=0, e=0, f=1, g=0),
            0b1010010
        )

    def test_no_fields(self):
        self.assertEqual(
            pyrtl.bitpattern_to_val('1010010'),
            0b1010010
        )

    def test_error_both_ordered_and_named_fields(self):
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.bitpattern_to_val('iiiiiiirrrrrsssss010iiiii0100011', 1, r=3, s=4)

    def test_error_invalid_num_unique_patterns(self):
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.bitpattern_to_val('iiiiiiirrrrrsssss010iiiii0100011', 1, 3)
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.bitpattern_to_val('iiiiiiirrrrrsssss010iiiii0100011', i=1, r=3)

    def test_error_bitpattern_field_not_provided(self):
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.bitpattern_to_val('iiiiiiirrrrrsssss010iiiii0100011', i=1, r=3, t=4)

    def test_error_unnamed_fields_in_bitpattern(self):
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.bitpattern_to_val('iiiiiii?????sssss010iiiii0100011', 1, 3, 4)

    def test_error_value_doesnt_fit_in_field(self):
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.bitpattern_to_val('iiiiiiirrrrrsssss010iiiii0100011', 1, 65, 4)

    def test_error_empty_bitpattern(self):
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.bitpattern_to_val('')


class TestChop(unittest.TestCase):
    def setUp(self):
        random.seed(8492049)
        pyrtl.reset_working_block()

    def check_trace(self, correct_string):
        sim_trace = pyrtl.SimulationTrace()
        sim = pyrtl.Simulation(tracer=sim_trace)
        for i in range(8):
            sim.step({})
        output = six.StringIO()
        sim_trace.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), correct_string)

    def test_fields_mismatch(self):
        instr = pyrtl.WireVector(name='instr', bitwidth=32)
        with self.assertRaises(pyrtl.PyrtlError):
            o = pyrtl.chop(instr, 10, 10, 10)
        with self.assertRaises(pyrtl.PyrtlError):
            o = pyrtl.chop(instr, 10, 10, 14)
        with self.assertRaises(pyrtl.PyrtlError):
            o = pyrtl.chop(instr, 33)

    def test_wrong_input_types_fail(self):
        instr = pyrtl.WireVector(name='instr', bitwidth=32)
        x = pyrtl.WireVector(name='x', bitwidth=10)
        with self.assertRaises(pyrtl.PyrtlError):
            o = pyrtl.chop(instr, x, 10, 12)
        with self.assertRaises(pyrtl.PyrtlError):
            o = pyrtl.chop(instr, 10, x, 12)
        with self.assertRaises(pyrtl.PyrtlError):
            o = pyrtl.chop(instr, x)
        with self.assertRaises(pyrtl.PyrtlError):
            o = pyrtl.chop(10, 5, 5)

    def test_chop_does_simulation_correct(self):
        r = pyrtl.Register(5, 'r')
        r.next <<= r + 1
        a, b, c = pyrtl.helperfuncs.chop(r, 2, 2, 1)
        o = pyrtl.Output(name='o')
        o <<= c
        self.check_trace('o 01010101\nr 01234567\n')


class TestBitField_Update(unittest.TestCase):
    def setUp(self):
        random.seed(8492049)
        pyrtl.reset_working_block()

    def check_trace(self, correct_string):
        sim_trace = pyrtl.SimulationTrace()
        sim = pyrtl.Simulation(tracer=sim_trace)
        for i in range(8):
            sim.step({})
        output = six.StringIO()
        sim_trace.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), correct_string)

    def test_field_too_big(self):
        a = pyrtl.WireVector(name='a', bitwidth=3)
        b = pyrtl.WireVector(name='b', bitwidth=3)
        with self.assertRaises(pyrtl.PyrtlError):
            o = pyrtl.bitfield_update(a, 1, 2, b)

    def test_field_too_big_truncate(self):
        a = pyrtl.WireVector(name='a', bitwidth=3)
        b = pyrtl.WireVector(name='b', bitwidth=3)
        o = pyrtl.bitfield_update(a, 1, 2, b, truncating=True)

    def test_no_bits_to_update(self):
        a = pyrtl.WireVector(name='a', bitwidth=3)
        b = pyrtl.WireVector(name='b', bitwidth=3)
        with self.assertRaises(pyrtl.PyrtlError):
            o = pyrtl.bitfield_update(a, 1, 1, b, truncating=True)

    def bitfield_update_checker(self, input_width, range_start, range_end,
                                update_width, test_amt=20):
        def ref(i, s, e, u):
            mask = ((1 << (e)) - 1) - ((1 << s) - 1)
            return (i & ~mask) | ((u << s) & mask)
        inp, inp_vals = utils.an_input_and_vals(input_width, test_vals=test_amt, name='inp')
        upd, upd_vals = utils.an_input_and_vals(update_width, test_vals=test_amt, name='upd')
        out = pyrtl.Output(input_width, "out")
        bfu_out = pyrtl.bitfield_update(inp, range_start, range_end, upd)
        self.assertEqual(len(out), len(bfu_out))  # output should have width of input
        out <<= bfu_out
        true_result = [ref(i, range_start, range_end, u) for i, u in zip(inp_vals, upd_vals)]
        upd_result = utils.sim_and_ret_out(out, [inp, upd], [inp_vals, upd_vals])
        self.assertEqual(upd_result, true_result)

    def test_bitfield(self):
        self.bitfield_update_checker(10, 3, 6, 3)

    def test_bitfield_msb(self):
        a = pyrtl.Const(0, 8)
        pyrtl.probe(pyrtl.bitfield_update(a, -1, None, 1), 'out')
        sim = pyrtl.Simulation()
        sim.step({})
        self.assertEqual(sim.inspect('out'), 0b10000000)

    def test_bitfield_all_but_msb(self):
        a = pyrtl.Const(0, 8)
        pyrtl.probe(pyrtl.bitfield_update(a, None, -1, 0b101110), 'out')
        sim = pyrtl.Simulation()
        sim.step({})
        self.assertEqual(sim.inspect('out'), 0b0101110)

    def test_bitfield_lsb(self):
        a = pyrtl.Const(0, 8)
        pyrtl.probe(pyrtl.bitfield_update(a, None, 1, 1), 'out')
        sim = pyrtl.Simulation()
        sim.step({})
        self.assertEqual(sim.inspect('out'), 0b00000001)

    def test_bitfield_all_but_lsb(self):
        a = pyrtl.Const(0, 8)
        pyrtl.probe(pyrtl.bitfield_update(a, 1, None, 0b1011101), 'out')  # TODO
        sim = pyrtl.Simulation()
        sim.step({})
        self.assertEqual(sim.inspect('out'), 0b10111010)


class TestBitField_Update_Set(unittest.TestCase):
    def setUp(self):
        random.seed(8492049)
        pyrtl.reset_working_block()

    def check_trace(self, correct_string):
        sim_trace = pyrtl.SimulationTrace()
        sim = pyrtl.Simulation(tracer=sim_trace)
        for i in range(8):
            sim.step({})
        output = six.StringIO()
        sim_trace.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), correct_string)

    def test_field_too_big(self):
        a = pyrtl.WireVector(name='a', bitwidth=3)
        b = pyrtl.WireVector(name='b', bitwidth=3)
        with self.assertRaises(pyrtl.PyrtlError):
            o = pyrtl.bitfield_update_set(a, {(1, 2): b})

    def test_field_too_big_truncate(self):
        a = pyrtl.WireVector(name='a', bitwidth=3)
        b = pyrtl.WireVector(name='b', bitwidth=3)
        o = pyrtl.bitfield_update_set(a, {(1, 2): b}, truncating=True)

    def test_no_bits_to_update(self):
        a = pyrtl.WireVector(name='a', bitwidth=3)
        b = pyrtl.WireVector(name='b', bitwidth=3)
        with self.assertRaises(pyrtl.PyrtlError):
            o = pyrtl.bitfield_update_set(a, {(1, 1): b}, truncating=True)

    def test_overlapping_ranges(self):
        a = pyrtl.WireVector(name='a', bitwidth=10)
        b = pyrtl.WireVector(name='b', bitwidth=10)
        c = pyrtl.WireVector(name='c', bitwidth=10)
        with self.assertRaises(pyrtl.PyrtlError):
            o = pyrtl.bitfield_update_set(a, {(1, 4): b, (3, 6): c}, truncating=True)
        with self.assertRaises(pyrtl.PyrtlError):
            o = pyrtl.bitfield_update_set(a, {(8, 10): b, (-1, None): c}, truncating=True)
        with self.assertRaises(pyrtl.PyrtlError):
            o = pyrtl.bitfield_update_set(a, {(None, 3): b, (0, 1): c}, truncating=True)

    def bitfield_update_set_checker(self, input_width, update_set_constraints, test_amt=20):
        from functools import reduce
        from collections import namedtuple
        Update_Info = namedtuple('Update_Info', ['range_start', 'range_end',
                                                 'update_wire', 'update_vals'])

        def ref(i, s, e, u):
            mask = ((1 << (e)) - 1) - ((1 << s) - 1)
            return (i & ~mask) | ((u << s) & mask)
        inp, inp_vals = utils.an_input_and_vals(input_width, test_vals=test_amt, name='inp')
        update_set_list = []
        for ix, (range_start, range_end, update_width) in enumerate(update_set_constraints):
            upd, upd_vals = utils.an_input_and_vals(update_width, test_vals=test_amt,
                                                    name='upd%d' % ix)
            update_set_list.append(Update_Info(range_start, range_end, upd, upd_vals))
        out = pyrtl.Output(input_width, "out")
        bfu_out = pyrtl.bitfield_update_set(
            inp,
            {(ui.range_start, ui.range_end): ui.update_wire for ui in update_set_list}
        )
        self.assertEqual(len(out), len(bfu_out))  # output should have width of input
        out <<= bfu_out
        true_result = []
        for ix, iv in enumerate(inp_vals):
            true_result.append(
                reduce(lambda w, ui: ref(w, ui.range_start, ui.range_end, ui.update_vals[ix]),
                       update_set_list, iv)
            )
        upd_result = utils.sim_and_ret_out(
            out,
            [inp] + [ui.update_wire for ui in update_set_list],
            [inp_vals] + [ui.update_vals for ui in update_set_list],
        )
        self.assertEqual(upd_result, true_result)

    def test_bitfield(self):
        self.bitfield_update_set_checker(10, [(3, 6, 3), (0, 2, 2)])


class TestAnyAll(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def check_trace(self, correct_string):
        sim_trace = pyrtl.SimulationTrace()
        sim = pyrtl.Simulation(tracer=sim_trace)
        for i in range(8):
            sim.step({})
        output = six.StringIO()
        sim_trace.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), correct_string)

    def test_any_only_on_1_bit_vectors(self):
        a = pyrtl.WireVector(name='a', bitwidth=3)
        b = pyrtl.WireVector(name='b', bitwidth=1)
        with self.assertRaises(pyrtl.PyrtlError):
            r = pyrtl.corecircuits.rtl_any(a, b)

    def test_all_only_on_1_bit_vectors(self):
        a = pyrtl.WireVector(name='a', bitwidth=1)
        b = pyrtl.WireVector(name='b', bitwidth=1)
        c = pyrtl.WireVector(name='c', bitwidth=3)
        with self.assertRaises(pyrtl.PyrtlError):
            r = pyrtl.corecircuits.rtl_all(a, b, c)

    def test_any_works_with_consts(self):
        a = pyrtl.WireVector(name='a', bitwidth=1)
        c = pyrtl.WireVector(name='c', bitwidth=1)
        r = pyrtl.corecircuits.rtl_any(a, 1, c)

    def test_all_works_with_consts(self):
        a = pyrtl.WireVector(name='a', bitwidth=1)
        c = pyrtl.WireVector(name='c', bitwidth=1)
        r = pyrtl.corecircuits.rtl_all(a, 1, c)

    def test_any_does_simulation_correct(self):
        r = pyrtl.Register(3, 'r')
        r.next <<= r + 1
        a, b, c = r[0], r[1], r[2]
        o = pyrtl.Output(name='o')
        o <<= pyrtl.corecircuits.rtl_any(a, b, c)
        self.check_trace('o 01111111\nr 01234567\n')

    def test_all_does_simulation_correct(self):
        r = pyrtl.Register(3, 'r')
        r.next <<= r + 1
        a, b, c = r[0], r[1], r[2]
        o = pyrtl.Output(name='o')
        o <<= pyrtl.corecircuits.rtl_all(a, b, c)
        self.check_trace('o 00000001\nr 01234567\n')


class TestTreeReduce(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_xor(self):
        wires, vals = utils.make_inputs_and_values(7, exact_bitwidth=8, dist=utils.uniform_dist)
        outwire = pyrtl.Output(name="test")

        import operator
        from six.moves import reduce
        outwire <<= pyrtl.tree_reduce(operator.xor, wires)

        out_vals = utils.sim_and_ret_out(outwire, wires, vals)
        true_result = [reduce(operator.xor, v) for v in zip(*vals)]
        self.assertEqual(out_vals, true_result)

    def test_empty(self):
        with self.assertRaises(pyrtl.PyrtlError):
            import operator
            pyrtl.tree_reduce(operator.add, [])


class TestXorAllBits(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def check_trace(self, correct_string):
        sim_trace = pyrtl.SimulationTrace()
        sim = pyrtl.Simulation(tracer=sim_trace)
        for i in range(8):
            sim.step({})
        output = six.StringIO()
        sim_trace.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), correct_string)

    def test_one_wirevector(self):
        r = pyrtl.Register(3, 'r')
        r.next <<= r + 1
        o = pyrtl.Output(name='o')
        o <<= pyrtl.corecircuits.xor_all_bits(r)
        self.check_trace('o 01101001\nr 01234567\n')

    def test_list_of_one_bit_wires(self):
        r = pyrtl.Register(2, 'r')
        r.next <<= r + 1
        o = pyrtl.Output(name='o')
        o <<= pyrtl.corecircuits.xor_all_bits([r[0], r[1]])
        self.check_trace('o 01100110\nr 01230123\n')

    def test_list_of_long_wires(self):
        in_wires, vals = utils.make_inputs_and_values(4, exact_bitwidth=13)
        out = pyrtl.Output(name='o')
        out <<= pyrtl.corecircuits.xor_all_bits(in_wires)
        expected = [v1 ^ v2 ^ v3 ^ v4 for v1, v2, v3, v4 in zip(*vals)]
        self.assertEqual(expected, utils.sim_and_ret_out(out, in_wires, vals))


class TestMux(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_mux_too_many_inputs(self):
        a = pyrtl.WireVector(name='a', bitwidth=3)
        b = pyrtl.WireVector(name='b', bitwidth=1)
        c = pyrtl.WireVector(name='c', bitwidth=1)
        s = pyrtl.WireVector(name='s', bitwidth=1)
        with self.assertRaises(pyrtl.PyrtlError):
            r = pyrtl.corecircuits.mux(s, a, b, c)

    def test_mux_not_enough_inputs(self):
        a = pyrtl.WireVector(name='a', bitwidth=3)
        b = pyrtl.WireVector(name='b', bitwidth=1)
        c = pyrtl.WireVector(name='c', bitwidth=1)
        s = pyrtl.WireVector(name='s', bitwidth=2)
        with self.assertRaises(pyrtl.PyrtlError):
            r = pyrtl.corecircuits.mux(s, a, b, c)

    def test_mux_not_enough_inputs_but_default(self):
        a = pyrtl.WireVector(name='a', bitwidth=3)
        b = pyrtl.WireVector(name='b', bitwidth=1)
        s = pyrtl.WireVector(name='s', bitwidth=2)
        r = pyrtl.corecircuits.mux(s, a, b, default=0)

    def test_mux_enough_inputs_with_default(self):
        a = pyrtl.WireVector(name='a', bitwidth=3)
        b = pyrtl.WireVector(name='b', bitwidth=1)
        c = pyrtl.WireVector(name='c', bitwidth=1)
        d = pyrtl.WireVector(name='d', bitwidth=1)
        s = pyrtl.WireVector(name='s', bitwidth=2)
        r = pyrtl.corecircuits.mux(s, a, b, c, d, default=0)

    def test_mux_too_many_inputs_with_default(self):
        a = pyrtl.WireVector(name='a', bitwidth=3)
        b = pyrtl.WireVector(name='b', bitwidth=1)
        c = pyrtl.WireVector(name='c', bitwidth=1)
        d = pyrtl.WireVector(name='d', bitwidth=1)
        e = pyrtl.WireVector(name='e', bitwidth=1)
        s = pyrtl.WireVector(name='s', bitwidth=2)
        with self.assertRaises(pyrtl.PyrtlError):
            r = pyrtl.corecircuits.mux(s, a, b, c, d, e, default=0)

    def test_mux_too_many_inputs_with_extra_kwarg(self):
        a = pyrtl.WireVector(name='a', bitwidth=3)
        b = pyrtl.WireVector(name='b', bitwidth=1)
        s = pyrtl.WireVector(name='s', bitwidth=2)
        with self.assertRaises(pyrtl.PyrtlError):
            r = pyrtl.corecircuits.mux(s, a, b, default=0, foo=1)


class TestMuxSimulation(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        random.seed(8492049)

    def setUp(self):
        pyrtl.reset_working_block()

    def test_simple_mux_1(self):
        self.mux_t_subprocess(4, 16)

    def test_simple_mux_2(self):
        self.mux_t_subprocess(6, 7)

    def mux_t_subprocess(self, addr_width, val_width):
        mux_ins, vals = utils.make_consts(num_wires=2**addr_width, exact_bitwidth=val_width)
        control, testctrl = utils.an_input_and_vals(addr_width, 40, "mux_ctrl")

        out = pyrtl.Output(val_width, "mux_out")
        out <<= pyrtl.corecircuits.mux(control, *mux_ins)

        true_result = [vals[i] for i in testctrl]
        mux_result = utils.sim_and_ret_out(out, (control,), (testctrl,))
        self.assertEqual(mux_result, true_result)

    def test_mux_with_default(self):
        addr_width = 5
        val_width = 9
        default_val = 170  # arbitrary value under 2**val_width
        num_defaults = 5
        mux_ins, vals = utils.make_consts(num_wires=2**addr_width - num_defaults,
                                          exact_bitwidth=val_width, random_dist=utils.uniform_dist)
        control, testctrl = utils.an_input_and_vals(addr_width, 40, "mux_ctrl", utils.uniform_dist)

        for i in range(5):
            vals.append(default_val)

        out = pyrtl.Output(val_width, "mux_out")
        out <<= pyrtl.corecircuits.mux(control, *mux_ins, default=pyrtl.Const(default_val))

        true_result = [vals[i] for i in testctrl]
        mux_result = utils.sim_and_ret_out(out, (control,), (testctrl,))
        self.assertEqual(mux_result, true_result)

    def test_select(self):
        vals = 12, 27
        mux_ins = [pyrtl.Const(x) for x in vals]
        control, testctrl = utils.an_input_and_vals(1, 40, "sel_ctrl", utils.uniform_dist)

        out = pyrtl.Output(5, "mux_out")
        out <<= pyrtl.corecircuits.select(control, falsecase=mux_ins[0], truecase=mux_ins[1])

        true_result = [vals[i] for i in testctrl]
        mux_result = utils.sim_and_ret_out(out, (control,), (testctrl,))
        self.assertEqual(mux_result, true_result)

    def test_select_no_pred(self):
        vals = 12, 27
        mux_ins = [pyrtl.Const(x) for x in vals]
        control, testctrl = utils.an_input_and_vals(1, 40, "sel_ctrl", utils.uniform_dist)

        out = pyrtl.Output(5, "mux_out")
        out <<= pyrtl.corecircuits.select(control, mux_ins[1], mux_ins[0])

        true_result = [vals[i] for i in testctrl]
        mux_result = utils.sim_and_ret_out(out, (control,), (testctrl,))
        self.assertEqual(mux_result, true_result)


class TestRtlProbe(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_bad_probe_wire(self):
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.probe(5)
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.probe('a')

    def test_simple_probe(self):
        i = pyrtl.Input(1)
        o = pyrtl.Output(1)
        o <<= pyrtl.probe(i + 1)

    def test_probe_wire(self):
        i = pyrtl.Input(1)
        x = pyrtl.probe(i)
        self.assertIs(x, i)

    def test_simple_probe_debug(self):
        pyrtl.set_debug_mode()
        i = pyrtl.Input(1)
        o = pyrtl.Output(1)
        output = six.StringIO()
        sys.stdout = output
        o <<= pyrtl.probe(i + 1, name="probe0")
        sys.stdout = sys.__stdout__
        self.assertTrue(output.getvalue().startswith("Probe: probe0"))
        pyrtl.set_debug_mode(False)


class TestShiftSimulation(unittest.TestCase):

    def setUp(self):
        random.seed(8492049)
        pyrtl.reset_working_block()

    def shift_checker(self, shift_func, ref_func, input_width, shift_width,
                      shift_amount=None, test_amt=20):
        inputs, all_inp_vals = [], []
        inp, inp_vals = utils.an_input_and_vals(input_width, test_vals=test_amt, name='inp')
        inputs.append(inp)
        all_inp_vals.append(inp_vals)
        if shift_amount is not None:
            shf, shf_vals = shift_amount, [shift_amount] * test_amt
        else:
            shf, shf_vals = utils.an_input_and_vals(shift_width, test_vals=test_amt, name='shf')
            inputs.append(shf)
            all_inp_vals.append(shf_vals)
        out = pyrtl.Output(input_width, "out")
        shf_out = shift_func(inp, shf)
        self.assertEqual(len(out), len(shf_out))  # output should have width of input
        out <<= shf_out
        true_result = [ref_func(i, s) for i, s in zip(inp_vals, shf_vals)]
        shift_result = utils.sim_and_ret_out(out, inputs, all_inp_vals)
        self.assertEqual(shift_result, true_result)

    def sll_checker(self, input_width, shift_width, shift_amount=None):
        mask = (1 << input_width) - 1

        def ref(i, s):
            return (i << s) & mask

        self.shift_checker(pyrtl.shift_left_logical, ref, input_width, shift_width, shift_amount)

    def sla_checker(self, input_width, shift_width, shift_amount=None):
        mask = (1 << input_width) - 1

        def ref(i, s):
            return (i << s) & mask

        self.shift_checker(pyrtl.shift_left_arithmetic, ref, input_width,
                           shift_width, shift_amount)

    def srl_checker(self, input_width, shift_width, shift_amount=None):
        mask = (1 << input_width) - 1

        def ref(i, s):
            return (i >> s) & mask

        self.shift_checker(pyrtl.shift_right_logical, ref, input_width, shift_width, shift_amount)

    def sra_checker(self, input_width, shift_width, shift_amount=None):
        # a little more work is required to take the positive number and treat it
        # as a twos complement number for the purpose of testing the shifter
        def ref(i, s):
            mask = (1 << input_width) - 1
            if (i >> input_width - 1) & 0x1 == 0x1:
                return ((~mask | i) >> s) & mask  # negative number
            else:
                return (i >> s) & mask  # positive number
        self.shift_checker(pyrtl.shift_right_arithmetic, ref, input_width,
                           shift_width, shift_amount)

    def test_sll(self):
        self.sll_checker(5, 2)

    def test_sla(self):
        self.sla_checker(5, 2)

    def test_srl(self):
        self.srl_checker(5, 2)

    def test_sra(self):
        self.sra_checker(5, 2)

    def test_sll_big(self):
        self.sll_checker(10, 3)

    def test_sla_big(self):
        self.sla_checker(10, 3)

    def test_srl_big(self):
        self.srl_checker(10, 3)

    def test_sra_big(self):
        self.sra_checker(10, 3)

    def test_sll_over(self):
        self.sll_checker(4, 4)

    def test_sla_over(self):
        self.sla_checker(4, 4)

    def test_srl_over(self):
        self.srl_checker(4, 4)

    def test_sra_over(self):
        self.sra_checker(4, 4)

    def test_sll_integer_shift_amount(self):
        self.sll_checker(5, 2, 1)

    def test_sla_integer_shift_amount(self):
        self.sla_checker(5, 2, 1)

    def test_srl_integer_shift_amount(self):
        self.srl_checker(5, 2, 1)

    def test_sra_integer_shift_amount(self):
        self.sra_checker(5, 2, 1)


class TestBasicMult(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def mult_t_base(self, len_a, len_b):
        # Creating the logic nets
        a, b = pyrtl.Input(len_a, "a"), pyrtl.Input(len_b, "b")
        product = pyrtl.Output(name="product")
        product <<= pyrtl.corecircuits._basic_mult(a, b)

        self.assertEqual(len(product), len_a + len_b)

        # creating the testing values and the correct results
        xvals = [int(random.uniform(0, 2 ** len_a - 1)) for i in range(20)]
        yvals = [int(random.uniform(0, 2 ** len_b - 1)) for i in range(20)]
        true_result = [i * j for i, j in zip(xvals, yvals)]

        # Setting up and running the tests
        sim_trace = pyrtl.SimulationTrace()
        sim = pyrtl.Simulation(tracer=sim_trace)
        for cycle in range(len(xvals)):
            sim.step({a: xvals[cycle], b: yvals[cycle]})

        # Extracting the values and verifying correctness
        multiplier_result = sim_trace.trace[product.name]
        self.assertEqual(multiplier_result, true_result)

    def test_mult_1(self):
        self.mult_t_base(1, 7)

    def test_mult_1_1(self):
        self.mult_t_base(2, 1)

    def test_mult_2(self):
        self.mult_t_base(5, 4)

    def test_mult_3(self):
        self.mult_t_base(5, 2)


class TestRtlAssert(unittest.TestCase):

    class RTLSampleException(Exception):
        pass

    def setUp(self):
        pyrtl.reset_working_block()

    def bad_rtl_assert(self, *args, **kwargs):
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.rtl_assert(*args, **kwargs)

    def test_bad_type(self):
        self.bad_rtl_assert(True, self.RTLSampleException())
        self.bad_rtl_assert(1, self.RTLSampleException())

    def test_wrong_len(self):
        w = pyrtl.Input(2)
        w2 = pyrtl.Input()

        self.bad_rtl_assert(w, self.RTLSampleException())
        self.bad_rtl_assert(w2, self.RTLSampleException())

    def test_invalid_exception_type(self):
        w = pyrtl.Input(1)

        self.bad_rtl_assert(w, 1)
        self.bad_rtl_assert(w, "")
        self.bad_rtl_assert(w, w)
        self.bad_rtl_assert(w, KeyError())

    def test_duplicate_assert(self):
        w = pyrtl.Input(1)
        pyrtl.rtl_assert(w, self.RTLSampleException())
        pyrtl.rtl_assert(w, self.RTLSampleException())

    def test_wire_from_another_block(self):
        w = pyrtl.Input(1)
        pyrtl.reset_working_block()
        self.bad_rtl_assert(w, self.RTLSampleException())

    def test_wire_outside_block(self):
        w = pyrtl.Input(1)
        block = pyrtl.working_block()
        block.wirevector_set.clear()
        self.bad_rtl_assert(w, self.RTLSampleException())

    def test_create_assert(self):
        w = pyrtl.WireVector(1)
        pyrtl.rtl_assert(w, self.RTLSampleException('testing rtl assert'))

    def test_assert_simulation(self):
        i = pyrtl.Input(1)
        o = pyrtl.rtl_assert(i, self.RTLSampleException('test assertion failed'))

        sim = pyrtl.Simulation()
        sim.step({i: 1})
        self.assertEqual(sim.inspect(o), 1)

        with self.assertRaises(self.RTLSampleException):
            sim.step({i: 0})

    def test_assert_fastsimulation(self):
        i = pyrtl.Input(1)
        o = pyrtl.rtl_assert(i, self.RTLSampleException('test assertion failed'))

        sim = pyrtl.FastSimulation()
        sim.step({i: 1})
        self.assertEqual(sim.inspect(o), 1)

        with self.assertRaises(self.RTLSampleException):
            sim.step({i: 0})


class TestLoopDetection(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        # Redirect stdout because we don't care to see the specific messages about
        # wires being deemed useless by optimization sent to stdout.
        f = open(os.devnull, 'w')
        sys.stdout = f

    def tearDown(self):
        sys.stdout.close()
        sys.stdout = sys.__stdout__

    def assert_no_loop(self):
        self.assertEqual(pyrtl.find_loop(), None)
        pyrtl.synthesize()
        self.assertEqual(pyrtl.find_loop(), None)
        pyrtl.optimize()
        self.assertEqual(pyrtl.find_loop(), None)

    def assert_has_loop(self):
        self.assertNotEqual(pyrtl.find_loop(), None)
        pyrtl.synthesize()
        self.assertNotEqual(pyrtl.find_loop(), None)
        pyrtl.optimize()
        self.assertNotEqual(pyrtl.find_loop(), None)

    def test_no_loop_1(self):
        ins = [pyrtl.Input(8) for i in range(8)]
        outs = [pyrtl.Output(8) for i in range(3)]
        x_1 = ins[0] & ins[1]
        x_2 = ins[3] & ins[4]
        x_3 = ins[0] | ins[4]
        x_4 = ins[6] ^ ins[7]
        x_5 = ~ ins[2]
        x_6 = ~ x_2
        x_7 = x_1 ^ x_3 & x_6
        outs[0] <<= x_4
        outs[2] <<= x_5 | x_7
        outs[1] <<= (x_1 & x_7) ^ x_3

        self.assert_no_loop()

    def test_loop_1(self):
        ins = [pyrtl.Input(8) for i in range(8)]
        outs = [pyrtl.Output(8) for i in range(3)]
        x_1 = ins[0] & ins[1]
        x_2 = ins[3] & ins[4]
        x_3 = ins[0] | ins[4]
        x_4 = ins[6] ^ ins[7]
        x_5 = ~ ins[2]
        x_6 = ~ x_2
        l_1 = pyrtl.WireVector(8)
        l_0 = x_6 & l_1
        l_1 <<= (l_0 | x_5) & x_2

        x_7 = x_1 ^ x_3 & l_0
        outs[0] <<= x_4
        outs[2] <<= x_5 | l_1
        outs[1] <<= (x_1 & x_7) ^ x_3

        self.assert_has_loop()

    def test_no_loop_special_ops(self):
        mem1 = pyrtl.MemBlock(4, 4)
        ins = [pyrtl.Input(4) for i in range(8)]
        outs = [pyrtl.Output(4) for i in range(3)]
        reg = pyrtl.Register(4)

        x_1 = ins[4] < reg
        x_2 = ins[1] * x_1
        x_3 = pyrtl.corecircuits.mux(x_1, ins[1], ins[2])
        # x_4 = pyrtl.as_wires(mem1[ins[6]])
        x_4 = mem1[ins[6]]
        x_5 = reg + ins[7]
        x_9 = pyrtl.as_wires(x_4)
        mem1[pyrtl.as_wires(x_4)] <<= x_9
        outs[0] <<= x_2 == x_1
        reg.next <<= x_5 & ins[1]
        outs[1] <<= reg
        outs[2] <<= pyrtl.corecircuits.concat(x_1, x_5[:7])

        self.assert_no_loop()

    def test_edge_case_1(self):
        in_1 = pyrtl.Input(10)
        in_2 = pyrtl.Input(9)
        fake_loop_wire = pyrtl.WireVector(1)
        comp_wire = pyrtl.corecircuits.concat(in_2[0:4], fake_loop_wire, in_2[4:9])
        r_wire = in_1 & comp_wire
        fake_loop_wire <<= r_wire[3]
        out = pyrtl.Output(10)
        out <<= fake_loop_wire

        # Yes, because we only check loops on a net level, this will still be
        # a loop pre synth
        self.assertNotEqual(pyrtl.find_loop(), None)
        pyrtl.synthesize()

        # Because synth separates the individual wires, it also resolves the loop
        self.assertEqual(pyrtl.find_loop(), None)
        pyrtl.optimize()
        self.assertEqual(pyrtl.find_loop(), None)

    def test_loop_2(self):
        in_1 = pyrtl.Input(10)
        in_2 = pyrtl.Input(9)
        fake_loop_wire = pyrtl.WireVector(1)
        # Note the slight difference from the last test case on the next line
        comp_wire = pyrtl.corecircuits.concat(in_2[0:6], fake_loop_wire, in_2[6:9])
        r_wire = in_1 & comp_wire
        fake_loop_wire <<= r_wire[3]
        out = pyrtl.Output(10)
        out <<= fake_loop_wire

        # It causes there to be a real loop
        self.assert_has_loop()

    def test_no_loop_reg_1(self):
        reg = pyrtl.Register(8)
        in_w = pyrtl.Input(8)
        res = reg + in_w
        reg.next <<= res
        self.assert_no_loop()


if __name__ == "__main__":
    unittest.main()



"""

``Example``::

    ``csprng_trivium``
    load, req = pyrtl.Input(1, 'load'), pyrtl.Input(1, 'req')
    ready, rand = pyrtl.Output(1, 'ready'), pyrtl.Output(128, 'rand')
    ready_out, rand_out = prngs.csprng_trivium(128, load, req)
    ready <<= ready_out
    rand <<= rand_out
    sim_trace = pyrtl.SimulationTrace()
    sim = pyrtl.Simulation(tracer=sim_trace)

    # seed once at the beginning
    sim.step({'load': 1, 'req': 0})
    while sim.value[ready] == 0:  # or loop 19 cycles
        sim.step({'load': 0, 'req': 0})

    sim.step({'load': 0, 'req': 1})
    while sim.value[ready] == 0:  # or loop 2 cycles
        sim.step({'load': 0, 'req': 0})

    print(sim.inspect(rand))
    sim_trace.render_trace(symbol_len=45, segment_size=5)

    ``prng_xoroshiro128``
    load, req = pyrtl.Input(1, 'load'), pyrtl.Input(1, 'req')
    ready, rand = pyrtl.Output(1, 'ready'), pyrtl.Output(128, 'rand')
    ready_out, rand_out = prngs.prng_xoroshiro128(128, load, req)
    ready <<= ready_out
    rand <<= rand_out
    sim_trace = pyrtl.SimulationTrace()
    sim = pyrtl.Simulation(tracer=sim_trace)

    sim.step({'load': 1, 'req': 0})  # seed once at the beginning
    sim.step({'load': 0, 'req': 1})
    while sim.value[ready] == 0:  # or loop 2 cycles
        sim.step({'load': 0, 'req': 0})

    print(sim.inspect(rand))
    sim_trace.render_trace(symbol_len=40, segment_size=1)

    ``prng_lfsr``
    load, req = pyrtl.Input(1, 'load'), pyrtl.Input(1, 'req')
    rand = pyrtl.Output(64, 'rand')
    rand <<= prngs.prng_lfsr(64, load, req)
    sim_trace = pyrtl.SimulationTrace()
    sim = pyrtl.Simulation(tracer=sim_trace)

    sim.step({'load': 1, 'req': 0}) # seed once at the beginning
    sim.step({'load': 0, 'req': 1})
    sim.step({'load': 0, 'req': 0})
    print(sim.inspect(rand))
    sim_trace.render_trace(symbol_len=40, segment_size=1)

    ``explicit seeding``
    seed =  pyrtl.Input(127, 'seed')
    load, req = pyrtl.Input(1, 'load'), pyrtl.Input(1, 'req')
    rand = pyrtl.Output(32, 'rand')
    rand <<= prngs.prng_lfsr(32, load, req, seed)
    sim_trace = pyrtl.SimulationTrace()
    sim = pyrtl.Simulation(tracer=sim_trace)

    sim.step({'load': 1, 'req': 0, 'seed': 0x102030405060708090a0b0c0d0e0f010})
    sim.step({'load': 0, 'req': 1, 'seed': 0x102030405060708090a0b0c0d0e0f010})
    sim.step({'load': 0, 'req': 0, 'seed': 0x102030405060708090a0b0c0d0e0f010})
    print(sim.inspect(rand))
    sim_trace.render_trace(symbol_len=40, segment_size=1)

"""


from __future__ import absolute_import
import pyrtl


def prng_lfsr(bitwidth, load, req, seed=None):
    """ Builds a single-cycle PRNG using a 127 bits Fibonacci LFSR.

    :param bitwidth: the desired bitwidth of the random number
    :param load: one bit signal to load the seed into the prng
    :param req: one bit signal to request a random number
    :param seed: 127 bits WireVector, defaults to None (self-seeding),
      refrain from self-seeding if reseeding at run time is required
    :return: register containing the random number with the given bitwidth

    A very fast and compact PRNG that generates a random number using only one clock cycle.
    Has a period of 2**127 - 1. Its linearity makes it a bit statistically weak, but should be
    good enough for any noncryptographic purpose like test pattern generation.
    """
    # 127 bits is chosen because 127 is a mersenne prime, which makes the period of the
    # LFSR maximized at 2**127 - 1 for any requested bitwidth
    if seed is None:
        import random
        cryptogen = random.SystemRandom()
        seed = cryptogen.randrange(1, 2**127)  # seed itself if no seed signal is given

    lfsr = pyrtl.Register(127 if bitwidth < 127 else bitwidth)
    # leap ahead by shifting the LFSR bitwidth times
    leap_ahead = lfsr
    for i in range(bitwidth):
        leap_ahead = pyrtl.concat(leap_ahead, leap_ahead[125] ^ leap_ahead[126])

    with pyrtl.conditional_assignment:
        with load:
            lfsr.next |= seed
        with req:
            lfsr.next |= leap_ahead
    return lfsr[:bitwidth]


def prng_xoroshiro128(bitwidth, load, req, seed=None):
    """ Builds a PRNG using the Xoroshiro128+ algorithm in hardware.

    :param bitwidth: the desired bitwidth of the random number
    :param load: one bit signal to load the seed into the prng
    :param req: one bit signal to request a random number
    :param seed: 128 bits WireVector, defaults to None (self-seeding),
      refrain from self-seeding if reseeding at run time is required
    :return ready, rand: ready is a one bit signal showing the random number has been
      produced, rand is a register containing the random number with the given bitwidth

    An efficient noncryptographic PRNG, has much smaller area than Trivium.
    But it does require a 64-bit adder to compute the output, so it is a bit slower.
    Has a period of 2**128 - 1. Passes most statistical tests. Outputs a 64-bit random
    word each cycle, takes multiple cycles if more than 64 bits are requested, and MSBs
    of the random words are returned if the bitwidth is not a multiple of 64.

    See also http://xoroshiro.di.unimi.it/
    """
    from math import ceil, log
    from pyrtl.rtllib import adders
    from pyrtl.rtllib.libutils import _shifted_reg_next as shift  # for readability
    if seed is None:
        import random
        cryptogen = random.SystemRandom()
        seed = cryptogen.randrange(1, 2**128)  # seed itself if no seed signal is given
    seed = pyrtl.as_wires(seed, 128)

    s0, s1 = (pyrtl.Register(64) for i in range(2))
    output = pyrtl.WireVector(64)
    # update internal states by xoring, rotating, and shifting
    _s1 = s0 ^ s1
    s0_next = (shift(s0, 'l', 55) | shift(s0, 'r', 9)) ^ shift(_s1, 'l', 14) ^ _s1
    s1_next = shift(_s1, 'l', 36) | shift(_s1, 'r', 28)
    output <<= adders.kogge_stone(s0, s1)

    gen_cycles = int(ceil(bitwidth / 64))
    counter_bitwidth = int(ceil(log(gen_cycles, 2))) if gen_cycles > 1 else 1
    rand = pyrtl.Register(gen_cycles * 64)
    counter = pyrtl.Register(counter_bitwidth, 'counter')
    gen_done = counter == gen_cycles - 1
    state = pyrtl.Register(1)
    WAIT, GEN = (pyrtl.Const(x) for x in range(2))
    with pyrtl.conditional_assignment:
        with load:
            s0.next |= seed[:64]
            s1.next |= seed[64:]
            state.next |= WAIT
        with req:
            counter.next |= 0
            s0.next |= s0_next
            s1.next |= s1_next
            rand.next |= pyrtl.concat(rand, output)
            state.next |= GEN
        with state == GEN:
            with ~gen_done:
                counter.next |= counter + 1
                s0.next |= s0_next
                s1.next |= s1_next
                rand.next |= pyrtl.concat(rand, output)

    ready = ~load & ~req & (state == GEN) & gen_done
    return ready, rand[-bitwidth:]  # return MSBs because LSBs are less random


def csprng_trivium(bitwidth, load, req, seed=None, bits_per_cycle=64):
    """ Builds a cyptographically secure PRNG using the Trivium stream cipher.

    :param bitwidth: the desired bitwidth of the random number
    :param load: one bit signal to load the seed into the prng
    :param req: one bit signal to request a random number
    :param seed: 160 bits WireVector (80 bits key + 80 bits IV), defaults to None (self-seeding),
      refrain from self-seeding if reseeding at run time is needed
    :param bits_per_cycle: the number of output bits to generate in parallel each cycle,
      up to 64 bits, must be a power of two: either 1, 2, 4, 8, 16, 32, or 64
    :return ready, rand: ready is a one bit signal showing either the random number has
      been produced or the seed has been initialized, rand is a register containing the
      random number with the given bitwidth

    This prng uses Trivium's key stream as its random bits output.
    Both seed and key stream are MSB first (the earliest bit is stored at the MSB).
    Trivium has a seed initialization stage that discards the first weak 1152 output bits
    after each loading. Generation stage can take multiple cycles as well depending on the
    given bitwidth and bits_per_cycle.
    Has smaller gate area and faster speed than AES-CTR and any other stream cipher.
    Passes all known statistical tests. Can be used to generate encryption keys or IVs.
    Designed to securely generate up to 2**64 bits. If more than 2**64 bits is needed,
    must reseed after each generation of 2**64 bits.

    Trivium specifications:
    http://www.ecrypt.eu.org/stream/ciphers/trivium/trivium.pdf
    See also the eSTREAM portfolio page:
    http://www.ecrypt.eu.org/stream/e2-trivium.html
    """
    from math import ceil, log
    if (64 // bits_per_cycle) * bits_per_cycle != 64:
        raise pyrtl.PyrtlError('bits_per_cycle is invalid')
    if seed is None:
        import random
        cryptogen = random.SystemRandom()
        seed = cryptogen.randrange(2**160)  # seed itself if no seed signal is given
    seed = pyrtl.as_wires(seed, 160)
    key = seed[80:]
    iv = seed[:80]

    a = pyrtl.Register(93)
    b = pyrtl.Register(84)
    c = pyrtl.Register(111)
    feedback_a, feedback_b, feedback_c, output = ([] for i in range(4))
    for i in range(bits_per_cycle):
        t1 = a[65 - i] ^ a[92 - i]
        t2 = b[68 - i] ^ b[83 - i]
        t3 = c[65 - i] ^ c[110 - i]
        feedback_a.append(t3 ^ c[108 - i] & c[109 - i] ^ a[68 - i])
        feedback_b.append(t1 ^ a[90 - i] & a[91 - i] ^ b[77 - i])
        feedback_c.append(t2 ^ b[81 - i] & b[82 - i] ^ c[86 - i])
        output.append(t1 ^ t2 ^ t3)
    # update internal states by shifting bits_per_cycle times
    a_next = pyrtl.concat(a, *feedback_a)
    b_next = pyrtl.concat(b, *feedback_b)
    c_next = pyrtl.concat(c, *feedback_c)

    init_cycles = 1152 // bits_per_cycle
    gen_cycles = int(ceil(bitwidth / bits_per_cycle))
    counter_bitwidth = int(ceil(log(max(init_cycles + 1, gen_cycles), 2)))
    rand = pyrtl.Register(bitwidth)
    counter = pyrtl.Register(counter_bitwidth, 'counter')
    init_done = counter == init_cycles
    gen_done = counter == gen_cycles - 1
    state = pyrtl.Register(2)
    WAIT, INIT, GEN = (pyrtl.Const(x) for x in range(3))
    with pyrtl.conditional_assignment:
        with load:
            counter.next |= 0
            a.next |= key
            b.next |= iv
            c.next |= pyrtl.concat(pyrtl.Const("3'b111"), pyrtl.Const(0, 108))
            state.next |= INIT
        with req:
            counter.next |= 0
            a.next |= a_next
            b.next |= b_next
            c.next |= c_next
            rand.next |= pyrtl.concat(rand, *output)
            state.next |= GEN
        with state == INIT:
            with ~init_done:
                counter.next |= counter + 1
                a.next |= a_next
                b.next |= b_next
                c.next |= c_next
        with state == GEN:
            with ~gen_done:
                counter.next |= counter + 1
                a.next |= a_next
                b.next |= b_next
                c.next |= c_next
                rand.next |= pyrtl.concat(rand, *output)

    ready = ~load & ~req & ((state == INIT) & init_done | (state == GEN) & gen_done)
    return ready, rand
